"""
libocn exception handling
"""
#TODO: I feel like this is extremely clunky and and doens't really give very helpful errors.

import warnings

# Import status codes from _libocn_bindings
from ._libocn_bindings import (
    SUCCESS,
    OOB_ERROR,
    EROSION_FAILURE,
    NULL_POINTER_ERROR,
    SWAP_WARNING,
    MALFORMED_GRAPH_WARNING,
)

STATUS_CODES = {
    SUCCESS: "SUCCESS",
    EROSION_FAILURE: "EROSION_FAILURE",
    OOB_ERROR: "OOB_ERROR",
    NULL_POINTER_ERROR: "NULL_POINTER_ERROR",
    SWAP_WARNING: "SWAP_WARNING",
    MALFORMED_GRAPH_WARNING: "MALFORMED_GRAPH_WARNING",
}

class LibOCNError(Exception):
    """Base exception for libocn errors."""
    pass

class NullPointerError(LibOCNError):
    default_message = (
        "\nTrying to access a null pointer. "
        "Something has gone terribly wrong inside PyOCN or libocn. "
        "Please open an issue at https://github.com/alextsfox/PyOCN/issues."
    )

    def __init__(self, custom_message=None):
        self.custom_message = custom_message

    def __str__(self):
        if self.custom_message:
            return f"{self.default_message}\nDetails: {self.custom_message}"
        return self.default_message

class SwapWarning(RuntimeWarning):
    pass

class MalformedGraphWarning(RuntimeWarning):
    default_message = (
        "\nYour flowgrid is invalid. "
        "This can be due to either cycles in the graph, "
        "the root node not having access to all nodes, "
        "incorrect dimensions (must have even rows and columns), "
        "or other structural issues."
    )

    def __init__(self, custom_message=None):
        self.custom_message = custom_message

    def __str__(self):
        if self.custom_message:
            return f"{self.default_message}\nDetails: {self.custom_message}"
        return self.default_message

STATUS_EXCEPTION_MAP = {
    OOB_ERROR: IndexError,
    NULL_POINTER_ERROR: NullPointerError,
}

STATUS_WARNING_MAP = {
    SWAP_WARNING: SwapWarning,
    MALFORMED_GRAPH_WARNING: MalformedGraphWarning,
}
STATUS_OKAY_MAP = {
    EROSION_FAILURE: None, 
}
def check_status(status):
    """
    Checks the status code returned by libocn functions.
    Raises an exception or issues a warning as appropriate.
    """
    if status == SUCCESS:
        return
    if status in STATUS_EXCEPTION_MAP:
        raise STATUS_EXCEPTION_MAP[status](f"libocn error: {STATUS_CODES.get(status, status)} ({status})")
    elif status in STATUS_WARNING_MAP:
        warnings.warn(f"libocn warning: {STATUS_CODES.get(status, status)} ({status})", STATUS_WARNING_MAP[status])
    elif status in STATUS_OKAY_MAP:
        return
    else:
        raise LibOCNError(f"Unknown libocn status code: {status}")

__all__ = [
    "LibOCNError",
    "NullPointerError",
    "SwapWarning",
    "MalformedGraphWarning",
    "check_status",
]



import warnings
import ctypes
from typing import Any, Callable, TYPE_CHECKING
from os import PathLike
from numbers import Number
from pathlib import Path

import networkx as nx 
import numpy as np
import rasterio
from tqdm import tqdm

from ._statushandler import check_status
from .utils import simulated_annealing_schedule, net_type_to_dag, unwrap_digraph
from . import _libocn_bindings as _bindings
from . import _flowgrid_convert as fgconv

if TYPE_CHECKING:
    import xarray as xr

"""
High-level Optimized Channel Network (OCN) interface.

This module provides a high-level interface to the underlying
``libocn`` C library. The :class:`OCN` class can be used for 
constructing and optimizing river network models using 
simulated annealing.

Note
----
- The underlying data structure managed by :class:`OCN` is a FlowGrid owned by
    ``libocn``. Pointer lifetime and destruction are handled safely within this
    class.
- Many operations convert to a NetworkX ``DiGraph`` for convenience. These
    conversions are slow, and are intended for inspection, analysis, 
    and visualization rather than tight inner loops.

See Also
--------
PyOCN.utils
        Helper functions for OCN fitting and construction
PyOCN.plotting
        Helper functions for visualization and plotting
"""

# TODO: have to_rasterio use the option to set the root node to 0,0 by using to_xarray as the backend instead of numpy?




class OCN:
    """
    The main class for interacting with Optimized Channel Networks. 

    Use :meth:`OCN.from_net_type` or :meth:`OCN.from_digraph` to construct an
    instance. 
    
    Methods
    -------------------
    from_net_type
        Create an OCN from a predefined network type and dimensions.
    from_digraph
        Create an OCN from an existing NetworkX DiGraph.
    to_digraph
        Export the current grid to a NetworkX DiGraph.
    to_numpy
        Export raster arrays (energy, drained area, elevation) as numpy arrays.
    to_xarray
        Export raster arrays as an xarray Dataset (requires xarray).
    to_gtiff
        Export raster arrays to a GeoTIFF file (requires rasterio).
    copy
        Create a deep copy of the OCN.
    single_erosion_event
        Perform a single erosion event at a given temperature.
    fit
        Optimize the network using the simulated annealing method from Carraro et al (2020).
    fit_custom_cooling
        Optimize the network using a custom cooling function.
    compute_energy
        Compute the current energy of the network.
    copy
        Create a deep copy of the OCN.

    Attributes
    ----------
    energy : float
        Current energy of the network (read-only property).
    dims : tuple[int, int]
        Grid dimensions (rows, cols) of the FlowGrid (read-only property).
    resolution: float
        The side length of each grid cell (read-only property).
    nroots : int
        Number of root nodes in the current OCN grid (read-only property).
    gamma : float
        Exponent in the energy model.
    verbosity : int
        Verbosity level for underlying library output (0-2).
    wrap : bool
        If true, enables periodic boundary conditions on the FlowGrid (read-only property).
    history : np.ndarray
        numpy array of shape (n_iterations, 3) recording the iteration index, energy, and temperature at each iteration during optimization.
        Updated each time an optimization method is called.
    rng : int
        the current random state of the internal RNG
    vertical_exaggeration : float
        Vertical exaggeration factor for the network. Used when computing elevation.
    

    Examples
    --------
    The following is a simple example of creating, optimizing, and plotting
    an OCN using PyOCN and Matplotlib. More examples are available in the
    `demo.ipynb` notebook in the repository (https://github.com/alextsfox/PyOCN).

    >>> # Fit an OCN from a "Hip-roof" initial network shape and periodic boundary conditions
    >>> import matplotlib.pyplot as plt
    >>> import matplotlib as mpl
    >>> import PyOCN as po
    >>> ocn = po.OCN.from_net_type("H", dims=(64, 64), wrap=True, random_state=8472, verbosity=0)
    >>> ocn.fit()
    >>> po.plot_ocn_raster(ocn, norm=mpl.colors.PowerNorm(gamma=0.5), attribute='drained_area')
    >>> plt.show()
    """

    ###########################
    #    CONSTRUCTORS         #
    ###########################
    def __init__(self, dag: nx.DiGraph, resolution: float=1.0, gamma: float = 0.5, random_state=None, verbosity: int = 0, validate:bool=True, wrap : bool = False, vertical_exaggeration: float = 1.0):
        """
        Construct an :class:`OCN` from a valid NetworkX ``DiGraph``.

        Attention
        ---------
        Please use the classmethods :meth:`OCN.from_net_type` or
        :meth:`OCN.from_digraph` to instantiate an OCN.
        """
        
        # validate gamma, annealing schedule, and random_state
        if not isinstance(gamma, Number):
            raise TypeError(f"gamma must be a scalar. Got {type(gamma)}.")
        if not (0 <= gamma <= 1):
            warnings.warn(f"gamma values outside of [0, 1] may not be physically meaningful. Got {gamma}.")
        if not isinstance(resolution, Number):
            raise TypeError(f"resolution must be numeric. Got {type(resolution)}")
        
        self.verbosity = verbosity
        self.gamma = gamma
        self.__p_c_graph = fgconv.from_digraph(  # does most of the work 
            dag, 
            resolution, 
            verbose=(verbosity > 1), 
            validate=validate, 
            wrap=wrap
        )
        self.rng = random_state
        self.__p_c_graph.contents.energy = self.compute_energy()

        self.__history = np.empty((0, 3), dtype=np.float64)

        if not (isinstance(vertical_exaggeration, Number) and vertical_exaggeration >= 0):
            raise TypeError(f"vertical_exaggeration must be a number >= 0. Got {type(vertical_exaggeration)}")
        self.vertical_exaggeration = vertical_exaggeration


    @classmethod
    def from_net_type(cls, net_type:str, dims:tuple[int, int], resolution:float=1, gamma : float = 0.5, random_state=None, verbosity:int=0, wrap : bool = False, vertical_exaggeration: float = 1.0):
        """
        Create an :class:`OCN` from a predefined network type and dimensions.

        Parameters
        ----------
        net_type : str
            Predefined network type to instantiate from. Valid types are "H", "I", "E", and "V". See
            :func:`~PyOCN.utils.net_type_to_dag` for more information.
        dims : tuple[int, int]
            Grid dimensions (rows, cols). Both must be positive even integers.
        resolution : int, optional
            The side length of each grid cell.
        gamma : float, default 0.5
            Exponent in the energy model.
        random_state : int | numpy.random.Generator | None, optional
            Seed or generator for RNG seeding.
        verbosity : int, default 0
            Verbosity level (0-2) for underlying library output.
        wrap : bool, default False
            If true, allows wrapping around the edges of the grid (toroidal). If false, no wrapping is applied.
        vertical_exaggeration : float, default 1.0
            Vertical exaggeration factor for the network. Used when computing elevation.

        Returns
        -------
        OCN
            A newly constructed instance initialized from the specified
            network type and dimensions.
        """
        if not isinstance(dims, tuple):
            raise TypeError(f"dims must be a tuple of two positive integers, got {type(dims)}")
        if not (
            len(dims) == 2 
            and all(isinstance(d, int) and d > 0 for d in dims)
        ):
            raise ValueError(f"dims must be a tuple of two positive integers, got {dims}")
        
        if verbosity == 1:
            print(f"Creating {net_type} network DiGraph with dimensions {dims}...", end="")
        dag = net_type_to_dag(net_type, dims)
        if verbosity == 1:
            print(" Done.")
        
        # no need to validate inputs when using a predefined net_type. Saves time.
        return cls(dag, resolution, gamma, random_state, verbosity=verbosity, validate=False, wrap=wrap, vertical_exaggeration=vertical_exaggeration)

    @classmethod
    def from_digraph(cls, dag: nx.DiGraph, resolution:float=1, gamma=0.5, random_state=None, verbosity: int = 0, wrap: bool = False, vertical_exaggeration: float = 1.0):
        """
        Create an :class:`OCN` from an existing NetworkX ``DiGraph``.

        Parameters
        ----------
        dag : nx.DiGraph
            Directed acyclic graph (DAG) representing the stream network.
        resolution : int, optional
            The side length of each grid cell.
        gamma : float, default 0.5
            Exponent in the energy model.
        random_state : int | numpy.random.Generator | None, optional
            Seed or generator for RNG seeding.
        verbosity : int, default 0
            Verbosity level (0-2) for underlying library output.
        wrap : bool, default False
            If true, allows wrapping around the edges of the grid (toroidal). If false, no wrapping is applied.
        vertical_exaggeration : float, default 1.0
            Vertical exaggeration factor for the network. Used when computing elevation.

        Returns
        -------
        OCN
            A newly constructed instance encapsulating the provided graph.

        Important
        ---------
        The input graph must satisfy all of the following:

        - It is a directed acyclic graph (DAG).
        - Each node has attribute ``pos=(row:int, col:int)`` specifying its
          grid position with non-negative coordinates. Any other attributes
          are ignored.
        - The graph can be partitioned into one or more spanning trees over 
          a dense grid of shape ``(m, n)``: each grid cell corresponds to 
          exactly one node; each non-root node has ``out_degree == 1``; 
          the roots have ``out_degree == 0``.
        - Edges connect only to one of the 8 neighbors (cardinal or diagonal),
          i.e., no jumps over rows or columns. If ``wrap=True``, edges may
          connect across the grid boundaries (i.e. row 0 can connect to row m-1 and col 0 can connect to col n-1).
        - Edges do not cross in the row-column plane.
        - Both ``m`` and ``n`` are positive integers, and there are at least four
          vertices.

        Examples
        --------
        >>> # creating a "zig-zag" network
        >>> # O O O O
        >>> # |/|/|/|
        >>> # O O O X
        >>> import networkx as nx
        >>> import PyOCN as po
        >>> import matplotlib as mpl
        >>> dag = nx.DiGraph()
        >>> for i in range(8):
        ...     dag.add_node(i, pos=(i % 4, i // 4))
        ...     if i < 7: dag.add_edge(i, i + 1)
        >>> ocn = OCN.from_digraph(dag, random_state=8472)
        >>> ocn.fit()
        >>> po.plot_ocn_raster(ocn, norm=mpl.colors.PowerNorm(gamma=0.5), attribute='drained_area')
        >>> plt.show()
        """

        return cls(dag, resolution, gamma, random_state, verbosity=verbosity, validate=True, wrap=wrap, vertical_exaggeration=vertical_exaggeration)

    ###########################
    #    DUNDER METHODS       #
    ###########################
    def __repr__(self):
        return f"<PyOCN.OCN object at 0x{id(self):x}>\n<FlowGrid_C struct at 0x{ctypes.addressof(self.__p_c_graph.contents):x}>\n<Vertex_C array at 0x{ctypes.addressof(self.__p_c_graph.contents.vertices):x}>"
    def __str__(self):
        return f"OCN(gamma={self.gamma}, energy={self.energy}, dims={self.dims}, resolution={self.resolution}m, verbosity={self.verbosity})"
    def __del__(self):
        try:
            _bindings.libocn.fg_destroy(self.__p_c_graph)
            self.__p_c_graph = None
        except AttributeError:
            pass
    def __sizeof__(self) ->int:
        return (
            object.__sizeof__(self) +
            self.gamma.__sizeof__() +
            self.__history.nbytes +
            ctypes.sizeof(_bindings.FlowGrid_C) + 
            ctypes.sizeof(_bindings.Vertex_C)*(self.dims[0]*self.dims[1])
        )
    def __copy__(self) -> "OCN":
        """
        Create a deep copy of the OCN, including the underlying FlowGrid_C.
        Also copies the current RNG state. The new copy and the original
        will be independent from each other and behave identically statistically.

        If you want the copy to have a different random state, call :meth:`reseed`
        after copying.
        """
        cpy = object.__new__(type(self))
        cpy.gamma = self.gamma
        cpy.verbosity = self.verbosity
        
        cpy.__history = self.history.copy()

        cpy_p_c_graph = _bindings.libocn.fg_copy(self.__p_c_graph)
        if not cpy_p_c_graph:
            raise MemoryError("Failed to copy FlowGrid_C in OCN.__copy__")
        cpy.__p_c_graph = cpy_p_c_graph
        cpy.vertical_exaggeration = self.vertical_exaggeration
        return cpy

    def __deepcopy__(self, memo) -> "OCN":
        """
        Create a deep copy of the OCN, including the underlying FlowGrid_C.
        Also copies the current RNG state. The new copy and the original
        will be independent from each other and behave identically statistically.
        """
        return self.__copy__()
    
    def copy(self) -> "OCN":
        """
        Create a deep copy of the OCN, including the underlying FlowGrid_C.
        Also copies the current RNG state. The new copy and the original
        will be independent from each other and behave identically statistically.
        """
        return self.__copy__()

    ###########################
    #    PROPERTIES           #
    ###########################
    def compute_energy(self) -> float:
        """
        Compute the current energy of the network.

        Returns
        -------
        float
            The computed energy value.
        """
        return _bindings.libocn.ocn_compute_energy(self.__p_c_graph, self.gamma)
    
    @property
    def energy(self) -> float:
        return self.__p_c_graph.contents.energy
    @property
    def resolution(self) -> float:
        return self.__p_c_graph.contents.resolution
    @property
    def nroots(self) -> int:
        return int(self.__p_c_graph.contents.nroots)
    @property
    def dims(self) -> tuple[int, int]:
        return (
            int(self.__p_c_graph.contents.dims.row),
            int(self.__p_c_graph.contents.dims.col)
        )
    @property
    def wrap(self) -> bool:
        return self.__p_c_graph.contents.wrap
    @property
    def rng(self) -> int:
        s0, s1, s2, s3 = self.__p_c_graph.contents.rng.s
        s = (s0 << 96) | (s1 << 64) | (s2 << 32) | s3
        return s
    
    @rng.setter
    def rng(self, random_state:int|None|np.random.Generator=None):
        if not isinstance(random_state, (int, np.integer, type(None), np.random.Generator)):
            raise ValueError("RNG must be initialized with an integer/Generator/None.")
        seed = np.random.default_rng(random_state).integers(0, int(2**32 - 1))
        rng = _bindings.rng_state_t()
        _bindings.libocn.rng_seed(ctypes.byref(rng), seed)
        self.__p_c_graph.contents.rng = rng

    @property
    def history(self) -> np.ndarray:
        return self.__history

    ###########################
    #    GRAPH TRAVERSAL      #
    ###########################
    # # Commenting out for now. May decide to re-introduce later
    # def predecessors(self, pos:tuple[int, int]) -> Generator[tuple[int, int], None, None]:
    #     """Returns an iterator over predecessor nodes of the node at position `pos` in the OCN.
    #     A predecessor is defined as an immediate upstream neighbor.

    #     A predecessor of a node n is a node m such that there exists a directed edge from m to n.

    #     Mirrors the behavior of `networkx.DiGraph.predecessors`, but works directly on the OCN C graph structure.

    #     Yields (row, col) of successors.

    #     Parameters
    #     ----------
    #     ocn : OCN
    #         The OCN instance.
    #     pos : tuple[int, int]
    #         The (row, col) position of the node whose predecessors are to be found.
        
    #     Raises
    #     -------
    #     TypeError
    #         If `pos` is not a tuple of two integers.
    #     IndexError
    #         If `pos` is out of bounds for the current OCN grid.

    #     See Also
    #     --------
    #     :meth:`OCN.successors`
    #     """

    #     pos = tuple(pos)
    #     if len(pos) != 2 or not all(isinstance(p, int) for p in pos):
    #         raise TypeError(f"Position must be a tuple of two integers. Got {pos}.")
    #     if (pos[0] < 0 or pos[0] >= self.dims[0]) or (pos[1] < 0 or pos[1] >= self.dims[1]):
    #         raise IndexError(f"Position {pos} is out of bounds for OCN with dimensions {self.dims}.")

    #     # convert (row, col) to linear index
    #     a = _bindings.libocn.fg_cart_to_lin(
    #         _bindings.CartPair_C(row=pos[0], col=pos[1]),
    #         _bindings.CartPair_C(row=self.dims[0], col=self.dims[1]),
    #     )

    #     upstream_indices = (_bindings.linidx_t * 8)()
    #     nupstream = _bindings.linidx_t()
    #     check_status(_bindings.libocn.fg_find_upstream_neighbors(
    #         upstream_indices, 
    #         ctypes.byref(nupstream),
    #         self.__p_c_graph, 
    #         _bindings.linidx_t(a)
    #     ))
    #     count = int(nupstream.value)

    #     # convert back to cartesian (row, col)
    #     for idx in upstream_indices[:count]:
    #         cart = _bindings.libocn.fg_lin_to_cart(
    #             idx, 
    #             _bindings.CartPair_C(row=self.dims[0], col=self.dims[1])
    #         )
    #         yield (int(cart.row), int(cart.col))


    # def successors(self, pos: tuple[int, int]) -> Generator[tuple[int, int], None, None]:
    #     """Returns an iterator over successor nodes of the node at position `pos` in the OCN.
    #     A successor is the immediate downstream neighbor (at most one; none if root).

    #     Mirrors the behavior of `networkx.DiGraph.successors`, but works directly on the OCN C graph structure.
    #     Yields (row, col) of successors.

    #     Parameters
    #     ----------
    #     pos : tuple[int, int]
    #         The (row, col) position of the node whose successors are to be found.
        
    #     Raises
    #     -------
    #     TypeError
    #         If `pos` is not a tuple of two integers.
    #     IndexError
    #         If `pos` is out of bounds for the current OCN grid.
    #     """
    #     pos = tuple(pos)
    #     if len(pos) != 2 or not all(isinstance(p, int) for p in pos):
    #         raise TypeError(f"Position must be a tuple of two integers. Got {pos}.")
    #     if (pos[0] < 0 or pos[0] >= self.dims[0]) or (pos[1] < 0 or pos[1] >= self.dims[1]):
    #         raise IndexError(f"Position {pos} is out of bounds for OCN with dimensions {self.dims}.")

    #     # (row, col) -> linear index
    #     a = _bindings.libocn.fg_cart_to_lin(
    #         _bindings.CartPair_C(row=pos[0], col=pos[1]),
    #         _bindings.CartPair_C(row=self.dims[0], col=self.dims[1]),
    #     )

    #     # Read vertex to get downstream direction
    #     vert = _bindings.Vertex_C()
    #     check_status(_bindings.libocn.fg_get_lin(
    #         ctypes.byref(vert),
    #         self.__p_c_graph,
    #         _bindings.linidx_t(a),
    #     ))

    #     # If root, no successors
    #     if int(vert.downstream) == _bindings.IS_ROOT:
    #         return
        
    #     cart = _bindings.libocn.fg_lin_to_cart(
    #         vert.adown, 
    #         _bindings.CartPair_C(row=self.dims[0], col=self.dims[1])
    #     )
    #     yield (int(cart.row), int(cart.col))        

    ###########################
    #    EXPORT               #
    ###########################
    def to_digraph(self) -> nx.DiGraph:
        """
        Create a NetworkX ``DiGraph`` view of the current grid.

        Returns
        -------
        nx.DiGraph
        # TODO: implement elevation, length, slope, q
            A DAG with the following node attributes per node:

            - ``pos``: ``(row, col)`` grid position
            - ``drained_area``: drained area value
            - ``energy``: cumulative energy at the node
            - ``elevation``: elevation at the node

            and the following edge attributes per edge:
            - ``length``: length of the edge/link
            - ``slope``: slope of the edge/link
            - ``q``: the landscape-forming discharge through the edge/link

        """
        dag = fgconv.to_digraph(self.__p_c_graph.contents, self.gamma, self.vertical_exaggeration)

        node_energies = dict()

        for node in nx.topological_sort(dag):
            node_energies[node] = (
                dag.nodes[node]['drained_area']**self.gamma 
                + sum(node_energies[p] for p in dag.predecessors(node))
            )
        nx.set_node_attributes(dag, node_energies, 'energy')

        return dag
    
    def to_gtiff(self, west:float, north:float, crs: Any, path:"str|PathLike|rasterio.io.MemoryFile", unwrap:bool=True):
        """
        Export a raster of the current FlowGrid to a GeoTIFF file
        using rasterio. The resulting raster has 3 bands: `energy`, `drained_area`, and `elevation`.

        Caution
        -------
        This uses the :attr:`resolution` attribute to set pixel size in the raster and for computing elevation.
        This function does not check for unit compatibility, so it is up to the user
        to ensure the resolution and CRS units match. By default,
        the resolution is set to 1.0 units. Using a lat-long (geographic) CRS may give undesireable results.
        It is recommended to use a linear (projected) CRS with units of length (e.g., meters) for best results.

        Parameters
        ----------
        west : float
            The western border of the raster in crs units, corresponding
            to column 0.
        north : float
            The northern border of the raster in crs units, corresponding
            to row 0.
        crs : Any
            The crs for the resulting gtiff, passed to `rasterio.open`
        path : str or Pathlike or rasterio.io.MemoryFile
            The output path for the resulting gtiff file. Can also be rasterio.io.MemoryFile. 
        unwrap : bool, default True
            If True and the current OCN has periodic boundaries, the
            resulting raster will be transformed so connected grid cells 
            are adjacent in the raster. This will result in a larger raster
            with some nan values. If False or the current OCN does not have
            periodic boundaries, then no transformation is applied and the
            raster will have the same dimensions as the current OCN grid.
        """
        try:
            import rasterio
            from rasterio.transform import from_origin
        except ImportError as e:
            raise ImportError(
                "PyOCN.OCN.to_gtiff() requires rasterio to be installed. Install with `pip install rasterio`."
            ) from e

        array = self.to_numpy(unwrap=unwrap)
        dims = array.shape[1:]
        energy = array[0]
        drained_area = array[1]
        elevation = array[2]

        transform = from_origin(west, north, self.resolution, self.resolution)

        # Write three bands: 1=energy, 2=drained_area, 3=elevation
        with rasterio.open(
            path,
            "w",
            driver="GTiff",
            height=dims[0],
            width=dims[1],
            count=3,
            dtype=np.float64,
            crs=crs,
            transform=transform,
            compress="deflate",
        ) as dst:
            dst.write(energy, 1)
            dst.write(drained_area, 2)
            dst.write(elevation, 3)
            # Band descriptions (nice to have)
            try:
                dst.set_band_description(1, "energy")
                dst.set_band_description(2, "drained_area")
                dst.set_band_description(3, "elevation")
            except Exception:
                pass
    
    def to_numpy(self, unwrap:bool=True) -> np.ndarray:
        """
        Export the current FlowGrid to a numpy array with shape (3, rows, cols).
        Has three channels: 0=energy, 1=drained_area, 2=elevation.

        Parameters
        ----------
        unwrap : bool, default True
            If True and the current OCN has periodic boundaries, the
            resulting array will be transformed so connected grid cells 
            are adjacent in the array. This will result in a larger array
            with some nan values. If False or the current OCN does not have
            periodic boundaries, then no transformation is applied and the
            resulting array will have the same dimensions as the current OCN grid.
        """
        dag = self.to_digraph()
        dims = self.dims
        if self.wrap and unwrap:
            dag = unwrap_digraph(dag, dims)
            positions = np.array(list(nx.get_node_attributes(dag, 'pos').values()))
            max_r, max_c = positions.max(axis=0)
            dims = (max_r + 1, max_c + 1)

        energy = np.full(dims, np.nan)
        drained_area = np.full(dims, np.nan)
        elevation = np.full(dims, np.nan)
        for node in dag.nodes:
            r, c = dag.nodes[node]['pos']
            energy[r, c] = dag.nodes[node]['energy']
            drained_area[r, c] = dag.nodes[node]['drained_area']
            elevation[r, c] = dag.nodes[node]['elevation']
        return np.stack([energy, drained_area, elevation], axis=0)

    def to_xarray(self, unwrap:bool=True) -> "xr.Dataset":
        """
        Export the current FlowGrid to an xarray Dataset
        
        Parameters
        ----------
        unwrap : bool, default True
            If True and the current OCN has periodic boundaries, the
            resulting rasters will be transformed so connected grid cells
            are adjacent in the output. This will result in a larger raster
            with some nan values. If False or the current OCN does not have
            periodic boundaries, then no transformation is applied and the
            resulting raster will have the same dimensions as the current OCN grid.
            
            When unwrapping, the (0,0) coordinate
            will be set to the position of the "main" root node, defined as
            the root node with the smallest row*cols + col value. Otherwise,
            (0,0) will be the top-left corner of the grid.
        
        Returns
        -------
        xr.Dataset
         an xarray Dataset with data variables:
            - `energy` (np.float64) representing energy at each grid cell
            - `area` (np.float64) representing drained area at each grid cell
            - `elevation` (np.float64) representing elevation at each grid cell
        and coordinates:
            - `y` (float) representing the northing coordinate of each row.
            - `x` (float) representing the easting coordinate of each column.
        """

        try:
            import xarray as xr
        except ImportError as e:
            raise ImportError(
                "PyOCN.OCN.to_xarray() requires xarray to be installed. Install with `pip install xarray`."
            ) from e
        
        dims = self.dims

        dag = self.to_digraph()
        row_root, col_root = 0, 0
        if self.wrap and unwrap:
            roots = [n for n, d in dag.out_degree() if d==0]
            main_root = min(roots, key=lambda n: dag.nodes[n]['pos'][0]*dims[1] + dag.nodes[n]['pos'][1])
            
            dag = unwrap_digraph(dag, dims)
            positions = np.array(list(nx.get_node_attributes(dag, 'pos').values()))
            max_r, max_c = positions.max(axis=0)
            dims = (max_r + 1, max_c + 1)
            
            # compute the new position of the root node after unwrapping. This will be the new origin (0,0).
            row_root, col_root = dag.nodes[main_root]['pos']

        energy = np.full(dims, np.nan)
        drained_area = np.full(dims, np.nan)
        elevation = np.full(dims, np.nan)
        for node in dag.nodes:
            r, c = dag.nodes[node]['pos']
            energy[r, c] = dag.nodes[node]['energy']
            drained_area[r, c] = dag.nodes[node]['drained_area']
            elevation[r, c] = dag.nodes[node]['elevation']
        array_out = np.stack([energy, drained_area, elevation], axis=0)

        dims = array_out.shape[1:]

        return xr.Dataset(
            data_vars={
                "energy": (["y", "x"], array_out[0].astype(np.float64)),
                "area": (["y", "x"], array_out[1].astype(np.float64)),
                "elevation": (["y", "x"], array_out[2].astype(np.float64)),
            },
            coords={
                "y": ("y", np.linspace(-row_root, (-row_root + (dims[0]-1)), dims[0])*self.resolution),
                "x": ("x", np.linspace(-col_root, (-col_root + (dims[1]-1)), dims[1])*self.resolution),
            },
            attrs={
                "description": "OCN fit result arrays",
                "resolution": self.resolution,
                "gamma": self.gamma,
                "rng": self.rng,
                "wrap": self.wrap,
            }
        )

    ###########################
    #    OPTIMIZATION         #
    ###########################
    def single_iteration(self, temperature:float, array_report:bool=False, unwrap:bool=True, calculate_full_energy:bool=False) -> "xr.Dataset | None":
        """ 
        Perform a single iteration of the optimization algorithm at a given temperature. Updates the internal history attribute.
        See :meth:`fit` for details on the algorithm.

        Parameters
        ----------
        temperature : float
            Temperature parameter governing acceptance probability. Typical
            range is a fraction of ocn.energy.
        array_report : bool, default False
            If True (default), the returned result will be an xarray.Dataset.
            See :meth:`fit` for details. The returned object will have an iteration dimension of size 1.
            Requires xarray to be installed.
        unwrap : bool, default True
            If True and the current OCN has periodic boundaries, the
            resulting rasters will be transformed so connected grid cells
            are adjacent in the output. This will result in a larger raster
            with some nan values. If False or the current OCN does not have
            periodic boundaries, then no transformation is applied and the
            resulting raster will have the same dimensions as the current OCN grid.
        calculate_full_energy : bool, default False
            If True, the full energy of the graph is recalculated when considering
            the proposed change. If False, a more efficient incremental update is used.
            Small numerical differences may arise between the two methods due to floating point
            precision. If precision is of the utmost importance, set this to True, but note that
            this comes with a significant performance penalty.

        Returns
        -------
        xr.Dataset | None
            If ``array_report == True``, an xarray.Dataset containing the state of the FlowGrid
            after the iteration. See :meth:`to_xarray` for details. If ``array_report == False``, returns None.

        Raises
        ------
        LibOCNError
            If the underlying C routine reports an error status.
        """
        # FlowGrid *G, uint32_t *total_tries, double gamma, double temperature
        check_status(_bindings.libocn.ocn_single_erosion_event(
            self.__p_c_graph,
            self.gamma, 
            temperature,
            calculate_full_energy
        ))

        # append to history
        last_iteration = self.__history[-1, 0] if self.__history.shape[0] else 0
        history = np.array([last_iteration + 1, self.energy, temperature]).reshape((1, 3))
        self.__history = np.concatenate((self.__history, history), axis=0)
        
        if not array_report:
            return None

        ds = self.to_xarray(unwrap=unwrap)
        ds = ds.expand_dims({"iteration": [0]})

        return ds
    
    def fit(
        self,
        cooling_rate:float=1.0,
        constant_phase:float=0.0,
        n_iterations:int=None,
        pbar:bool=False,
        array_reports:int=0,
        tol:float=None,
        max_iterations_per_loop=10_000,
        unwrap:bool=True,
        calculate_full_energy:bool=False) -> "xr.Dataset | None":
        """
        Convenience function to optimize the OCN using the simulated annealing algorithm from Carraro et al (2020).
        For finer control over the optimization process, use :meth:`fit_custom_cooling` or use :meth:`single_erosion_event` in a loop.

        This performs ``n_iterations`` erosion events, accepting or rejecting
        proposals according to a temperature schedule defined by the annealing algorithm. 
        A proposal consists of changing the outflow direction of a randomly selected vertex. 
        The new outflow direction is chosen uniformly from the valid neighbors.
        A proposal is valid if it maintains a well-formed graph structure.

        The :attr:`history` attribute is updated in-place after optimization finishes.

        Parameters
        ----------
        cooling_rate : float, default 1.0
            Cooling rate parameter in the annealing algorithm. Typical range is 0.5-1.5.
            Higher values result in faster cooling and a greedier search.
            Lower values result in slower cooling and more thorough exploration of the solution space, but slower convergence and lower stability.
        constant_phase : float, default 0.0
            Amount of time to hold temeprature constant at the start of the optimization.
            This is a fraction of n_iterations, and must be in the range [0, 1].
            A value of 0.0 (default) means the temperature starts cooling immediately
            from the initial temperature. A value of 1.0 means the temperature is held
            constant for the entire optimization.
        n_iterations : int, optional
            Total number of iterations. Defaults to ``40 * rows * cols``.
            Always at least ``energy_reports * 10`` (this should only matter for
            extremely small grids, where ``rows * cols < 256``).
        pbar : bool, default True
            Whether to display a progress bar.
        array_reports : int, default 0
            Number of timepoints (approximately) at which to save the state of the FlowGrid.
            If 0 (default), returns None. If >0, returns an xarray.Dataset
            containing the state of the FlowGrid at approximately evenly spaced intervals
            throughout the optimization, including the initial and final states. Requires xarray to be installed. See notes on xarray output for details.
        tol : float, optional
            If provided, optimization will stop early if the relative change
            in energy between reports is less than `tol`. Must be positive.
            If None (default), no early stopping is performed.
            Recommended values are in the range 1e-4 to 1e-6.
        max_iterations_per_loop: int, optional
            If provided, the number of iterations steps to perform in each "chunk"
            of optimization. Energy and output arrays can be reported no more often
            than this. Recommended values are 1_000-1_000_000. Default is 10_000.
        unwrap: bool, default True
            If True and the current OCN has periodic boundaries, the
            resulting rasters will be transformed so connected grid cells
            are adjacent in the output. This will result in a larger raster
            with some nan values. If False or the current OCN does not have
            periodic boundaries, then no transformation is applied and the
            resulting raster will have the same dimensions as the current OCN grid.
        calculate_full_energy : bool, default False
            If True, the full energy of the graph is recalculated when considering
            the proposed change. If False, a more efficient incremental update is used.
            Small numerical differences may arise between the two methods due to floating point
            precision. If precision is of the utmost importance, set this to True, but note that
            this comes with a significant performance penalty.

        Returns
        -------
        ds : xr.Dataset | None
            If ``array_reports > 0``, an xarray.Dataset containing the state of the FlowGrid
            at approximately evenly spaced intervals throughout the optimization, including
            the initial and final states. See notes on xarray output for details.
            If ``array_reports == 0``, returns None.

        

        Note
        ----
        The returned xarray.Dataset will have coordinates:

        - `y` (float) representing the northing coordinate of each row
        - `x` (float) representing the easting coordinate of each column
        - `iteration` (int) representing the iteration index at which the data was recorded

        and data variables:

        - `energy` (np.float64) representing energy at each grid cell
        - `area` (np.float64) representing drained area at each grid cell
        - `elevation` (np.float64) representing elevation at each grid cell

        If the OCN has a periodic boundary condition, the following changes apply: 

        - The (0,0) coordinate will be set to the position of the "main" root node, defined as the root node with the smallest row*cols + col value
        - The rasters will be unwrapped to a non-periodic representation, which may result in larger rasters.
        - The size of the final rasters are the maximum extent of the unwrapped grid, taken across all iterations.

        Generating reports requires additional memory and computation time.
        
        Note
        ----
        At iteration ``i``, the outflow of a random grid cell if proposed to be rerouted.
        The proposal is accepted with the probability
        
        .. math::
            P(\\text{accept}) = e^{-\Delta E / T},

        where :math:`\Delta E` is the change in energy the change would cause 
        and :math:`T` is the temperature of the network.

        The total energy of the system is computed from the drained areas of each grid cell :math:`k` as

        .. math::
            E = \sum_k A_k^\gamma

        The temperature of the network is governed by a cooling schedule, which is a function of iteration index.
        
        Note that when :math:`\Delta E < 0`, the move is always accepted.

        The cooling schedule used by this method is a piecewise function of iteration index:
        
        .. math::
            T(i) = \\begin{cases}
                E_0 & i < C N \\
                E_0 \cdot e^{\;i - C N} & i \ge C N
            \end{cases}

        where :math:`E_0` is the initial energy, :math:`N` is the total number
        of iterations, and :math:`C` is ``constant_phase``. Decreasing-energy
        moves (:math:`\Delta E < 0`) are always accepted.

        Alternative cooling schedules can be implemented using :meth:`fit_custom_cooling`.
        """
        
        # make sure energy is up to date, useful if the user modified any parameters manually
        self.__p_c_graph.contents.energy = self.compute_energy()
        if constant_phase is None:
            constant_phase = 0.0
        if cooling_rate is None:
            cooling_rate = 1.0
        if n_iterations is None:
            n_iterations = 40 * self.dims[0] * self.dims[1]

        # create a cooling schedule from arguments
        cooling_func = simulated_annealing_schedule(
            dims=self.dims,
            E0=self.energy,
            constant_phase=constant_phase,
            n_iterations=n_iterations,
            cooling_rate=cooling_rate,
        )

        return self.fit_custom_cooling(
            cooling_func=cooling_func,
            n_iterations=n_iterations,
            pbar=pbar,
            array_reports=array_reports,
            tol=tol,
            max_iterations_per_loop=max_iterations_per_loop,
            unwrap=unwrap,
            calculate_full_energy=calculate_full_energy,
        )

    def fit_custom_cooling(
        self,
        cooling_func:Callable[[np.ndarray], np.ndarray],
        n_iterations:int=None,
        iteration_start:int=0,
        pbar:bool=False,
        array_reports:int=0,
        tol:float=None,
        max_iterations_per_loop=10_000,
        unwrap:bool=True,
        calculate_full_energy:bool=False,
    ) -> "xr.Dataset | None":
        """
        Optimize the OCN using the a custom cooling schedule. This allows for
        multi-stage optimizations or other custom cooling schedules not covered by the default simulated annealing schedule
        from Carraro et al (2020).

        See :meth:`fit` for additional details on the optimization algorithm and parameters.

        Parameters
        ----------
        cooling_func : Callable[[np.ndarray], np.ndarray]
            A function that takes an array of iteration indices and returns an array of temperatures.
            This function defines the cooling schedule for the optimization. Note that the function
            should return temperatures that are appropriate for the current energy of the OCN.
        n_iterations : int, optional
        iteration_start : int, default 0
            The starting iteration index. This is useful for continuing an optimization
            from a previous run. Must be a non-negative integer. If provided, ``n_iterations``
            is the number of additional iterations to perform. The iteration number passed to
            ``cooling_func`` will be ``iteration_start + i`` where ``i`` is the current iteration index
            in the range ``[0, n_iterations-1]``.
        pbar : bool, default True
        array_reports : int, default 0
        tol : float, optional
        max_iterations_per_loop: int, optional
        unwrap: bool, default True
        calculate_full_energy: bool, default False

        Returns
        -------
        xr.Dataset | None
        """
        # validate inputs
        if n_iterations is None:
            n_iterations = int(40*self.dims[0]*self.dims[1])
        if not (isinstance(n_iterations, int) and n_iterations > 0):
            raise ValueError(f"n_iterations must be a positive integer, got {n_iterations}")
        if not (isinstance(array_reports, int) and array_reports >= 0):
            raise ValueError(f"array_reports must be a non-negative integer, got {array_reports}")
        if (not isinstance(iteration_start, (int, np.integer))) or iteration_start < 0:
            raise ValueError(f"iteration_start must be a non-negative integer, got {iteration_start}")
        if (not isinstance(tol, Number)) or tol < 0:
            if tol is not None:
                raise ValueError(f"tol must be a positive number or None, got {tol}")
        memory_est = array_reports*self.dims[0]*self.dims[1]*2*8 
        if memory_est > 20e6:
            warnings.warn(f"Requesting {array_reports} array is estimated to use {memory_est/1e6:.2f}MB of memory. Consider reducing array_reports or increasing max_iterations_per_loop if memory is a concern.")
        
        xarray_out = array_reports > 0
        if xarray_out:
            try:
                import xarray as xr
            except ImportError as e:
                raise ImportError(
                    "PyOCN.OCN.fit() with array_report>0 requires xarray to be installed. Install with `pip install xarray`."
                ) from e

        # preallocate output arrays
        max_iterations_per_loop = int(max_iterations_per_loop)
        max_iterations_per_loop = max(1, max_iterations_per_loop)
        n_iterations = int(n_iterations)
        n_iterations = max(1, n_iterations)
        max_iterations_per_loop = min(n_iterations, max_iterations_per_loop)

        # preallocate output arrays
        energy_out = np.empty(
            n_iterations//max_iterations_per_loop + 2, # always report energy when reporting arrays
            dtype=np.float64)
        
        if array_reports:
            array_report_interval = n_iterations // array_reports
            array_report_interval = max(array_report_interval, max_iterations_per_loop)
        else:
            array_report_interval = n_iterations*2  # never report arrays

        # temporary buffer to use for passing temperatures to libocn
        anneal_buf = np.empty(max_iterations_per_loop, dtype=np.float64)
        anneal_ptr = anneal_buf.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

        completed_iterations = iteration_start
        n_iterations += iteration_start

        # set up reporting
        array_report_idx = []
        energy_report_idx = []
        ds_out_dict = dict()
        array_report_idx.append(completed_iterations)
        if xarray_out:
            ds_out_dict[completed_iterations] = self.to_xarray(unwrap=self.wrap)
        energy_out[0] = self.energy
        energy_report_idx.append(completed_iterations)

        pbar = tqdm(
            total=n_iterations - iteration_start, 
            desc="OCN Fit", 
            unit_scale=True, 
            dynamic_ncols=True, 
            disable=not (pbar or self.verbosity >= 1)
        )
        
        while completed_iterations < n_iterations:
            iterations_this_loop = min(max_iterations_per_loop, n_iterations - completed_iterations)
            anneal_buf[:iterations_this_loop] = cooling_func(
                np.arange(completed_iterations, completed_iterations + iterations_this_loop)
            )

            # main call to optimizer in libocn
            e_old = self.energy
            check_status(_bindings.libocn.ocn_outer_ocn_loop(
                self.__p_c_graph, 
                iterations_this_loop, 
                self.gamma, 
                anneal_ptr,
                calculate_full_energy,
            ))
            e_new = self.energy
            completed_iterations += iterations_this_loop
            
            # always report energy
            energy_out[len(energy_report_idx)] = e_new
            energy_report_idx.append(completed_iterations)
            # report arrays if requested
            if (
                xarray_out 
                and (
                    (completed_iterations % array_report_interval) < max_iterations_per_loop  # intermediate report
                    or completed_iterations >= n_iterations  # final report
                )
            ):
                array_report_idx.append(completed_iterations)
                ds_out_dict[completed_iterations] = self.to_xarray(unwrap=unwrap)  # TODO: pre-allocate this dict?


            # progress bar update
            # TODO: move this to a separate function
            if pbar or self.verbosity >= 1:
                RED = '\033[31m'
                YELLOW = '\033[33m'
                CYAN = '\033[36m'
                END = '\033[0m'
                T_over_E = anneal_buf[iterations_this_loop - 1]/e_old*100
                ToE_str = f"{int(np.floor(T_over_E)):02d}.{int((T_over_E - np.floor(T_over_E))*100):02d}%"
                de_over_E = (e_new - e_old)/e_old*100

                dEoE_sign = '+' if de_over_E >= 0 else '-'
                dEoE_integer_part = int(np.floor(np.abs(de_over_E)))
                dEoE_fractional_part = int((np.abs(de_over_E) - dEoE_integer_part)*100)
                dEoE_str = f"{dEoE_sign}{dEoE_integer_part:02d}.{dEoE_fractional_part:02d}%"
                if T_over_E > 50: ToE_str = RED + ToE_str + END
                elif T_over_E > 5: ToE_str = YELLOW + ToE_str + END
                else: ToE_str = CYAN + ToE_str + END
                if de_over_E > 1: dEoE_str = RED + dEoE_str + END
                elif de_over_E > -1: dEoE_str = YELLOW + dEoE_str + END
                else: dEoE_str = CYAN + dEoE_str + END

                pbar.set_postfix({
                    "E": f"{self.energy:.1e}", 
                    "T/E": ToE_str,
                    "ΔE/E": dEoE_str,
                })
                pbar.update(iterations_this_loop)

            # check for convergence if requested
            if (
                (tol is not None) 
                and (e_new <= e_old) 
                and (abs((e_old - e_new)/e_old) if e_old > 0 else np.inf < tol)
            ):
                if self.verbosity > 1:
                    print("Convergence reached, stopping optimization.")
                break
        
        pbar.close()
        
        # update energy and temperature history
        last_history_iteration = int(self.history[-1, 0]) if self.history.shape[0] else 0
        skip_first = 1 if self.history.shape[0] else 0
        history = np.stack([
            last_history_iteration + np.asarray(energy_report_idx)[skip_first:] - iteration_start,
            energy_out[skip_first:len(energy_report_idx)],
            cooling_func(np.asarray(energy_report_idx[skip_first:])),
        ], axis=1).reshape(-1, 3)
        self.__history = np.concatenate([self.__history, history], axis=0)
        
        if not xarray_out: 
            return None
    
        # convert the ds_out_dict to a single xarray.Dataset with an iteration dimension if requested
        # TODO: move this to a separate function???
        
        # if unwrapping is requested, the output rasters may have different shapes, so we need to find the max extent
        # across all reported arrays and create a new raster with that shape.
        
        # find the maximum extent of the unwrapped grid across all reported arrays
        coord_ranges = list((ds.x.data.min(), ds.x.data.max(), ds.y.data.min(), ds.y.data.max()) for ds in ds_out_dict.values())
        xmin, xmax, ymin, ymax = (
            min(coord_range[0] for coord_range in coord_ranges), 
            max(coord_range[1] for coord_range in coord_ranges), 
            min(coord_range[2] for coord_range in coord_ranges), 
            max(coord_range[3] for coord_range in coord_ranges),
        )
        new_xcoords = np.arange(xmin, xmax + self.resolution, self.resolution)
        new_ycoords = np.arange(ymin, ymax + self.resolution, self.resolution)

        data_shape = (len(ds_out_dict), len(new_ycoords), len(new_xcoords))

        # build an empty dataset
        ds = xr.Dataset(
            data_vars={
                "energy": (
                    ["iteration", "y", "x"], 
                    np.full(data_shape, np.nan, dtype=np.float64)
                ),
                "area": (
                    ["iteration", "y", "x"], 
                    np.full(data_shape, np.nan, dtype=np.float64)
                ),
                "elevation": (
                    ["iteration", "y", "x"], 
                    np.full(data_shape, np.nan, dtype=np.float64)
                ),
            },
            coords={
                "iteration": ("iteration", np.asarray(array_report_idx)),
                "y": ("y", new_ycoords),
                "x": ("x", new_xcoords),
            },
            attrs={
                "description": "OCN fit result arrays",
                "resolution": self.resolution,
                "gamma": self.gamma,
                "rng": self.rng,
                "wrap": self.wrap,
            }
        )

        # fill in the dataset with the unwrapped arrays, matching coordinates
        for i, ds_i in ds_out_dict.items():
            ds.energy.loc[dict(iteration=i, y=ds_i.y, x=ds_i.x)] = ds_i.energy
            ds.area.loc[dict(iteration=i, y=ds_i.y, x=ds_i.x)] = ds_i.area
            ds.elevation.loc[dict(iteration=i, y=ds_i.y, x=ds_i.x)] = ds_i.elevation

        return ds



"""
Bindings for the libocn C library, 
providing ctypes-based access to C data structures and functions.
"""

from ctypes import (
    CDLL,
    Structure,
    c_uint16,
    c_uint32,
    c_double,
    c_uint8,
    c_bool,
    POINTER,
)
import importlib
from pathlib import Path

def _load_libocn():
    # attempt the dev build first
    _dev_lib = Path(__file__).parent / "libocn_dev.so"
    if _dev_lib.exists():
        return CDLL(str(_dev_lib))
    
    # prefer the built extension inside the wheel: PyOCN/_libocn*.so|.pyd
    try:
        mod = importlib.import_module("._libocn", package=__package__)
        return CDLL(str(Path(mod.__file__)))
    except Exception:
        pass

    raise OSError(
        "Could not locate libocn: neither the built extension (PyOCN._libocn) " \
        "nor a local shared library was found. Reinstall the wheel or build " \
        "the C library manually with PyOCN/c_src/build.sh."
    )

libocn = _load_libocn()




#############################
#   STATUS.H EQUIVALENTS    #
#############################
Status = c_uint8

SUCCESS = int(Status.in_dll(libocn, "SUCCESS").value)
EROSION_FAILURE = int(Status.in_dll(libocn, "EROSION_FAILURE").value)
OOB_ERROR = int(Status.in_dll(libocn, "OOB_ERROR").value)
NULL_POINTER_ERROR = int(Status.in_dll(libocn, "NULL_POINTER_ERROR").value)
SWAP_WARNING = int(Status.in_dll(libocn, "SWAP_WARNING").value)
MALFORMED_GRAPH_WARNING = int(Status.in_dll(libocn, "MALFORMED_GRAPH_WARNING").value)

########################
#  RNG.H EQUIVALENTS   #
########################

class rng_state_t(Structure):
    _fields_ = [
        ("s", c_uint32 * 4),
    ]

# void rng_seed(rng_state_t *rng, uint32_t seed);
libocn.rng_seed.argtypes = [POINTER(rng_state_t), c_uint32]
libocn.rng_seed.restype = None

# void rng_seed_random(rng_state_t *state);
libocn.rng_seed_random.argtypes = [POINTER(rng_state_t)]
libocn.rng_seed_random.restype = None

# uint32_t rng_randint32(rng_state_t *state);
libocn.rng_randint32.argtypes = [POINTER(rng_state_t)]
libocn.rng_randint32.restype = c_uint32

#############################
# STREAMGRAPH.H EQUIVALENTS #
#############################
drainedarea_t = c_double
cartidx_t = c_uint16

class CartPair_C(Structure):
    _fields_ = [
        ("row", cartidx_t),
        ("col", cartidx_t),
    ]
    
linidx_t = c_uint32
localedges_t = c_uint8
clockhand_t = c_uint8
IS_ROOT = int(clockhand_t.in_dll(libocn, "IS_ROOT").value)


class Vertex_C(Structure):
    _fields_ = [
        ("drained_area", drainedarea_t),
        ("adown", linidx_t),
        ("edges", localedges_t),
        ("downstream", clockhand_t),
        ("visited", c_uint8),
    ]

class FlowGrid_C(Structure):
    _fields_ = [
        ("dims", CartPair_C),
        ("energy", c_double),
        ("resolution", c_double),
        ("nroots", c_uint16),
        ("vertices", POINTER(Vertex_C)),  # Vertex*
        ("wrap", c_bool),
        ("rng", rng_state_t),  # rng_state_t
    ]

# linidx_t fg_cart_to_lin(CartPairC coords, CartPairC dims);
libocn.fg_cart_to_lin.argtypes = [CartPair_C, CartPair_C]
libocn.fg_cart_to_lin.restype = linidx_t

# CartPair fg_lin_to_cart(linidx_t a, CartPair dims);
libocn.fg_lin_to_cart.argtypes = [linidx_t, CartPair_C]
libocn.fg_lin_to_cart.restype = CartPair_C

# Status fg_clockhand_to_lin(linidx_t *a_down, linidx_t a, clockhand_t down, CartPair dims, bool wrap)
libocn.fg_clockhand_to_lin.argtypes = [POINTER(linidx_t), linidx_t, clockhand_t, CartPair_C, c_bool]
libocn.fg_clockhand_to_lin.restype = Status

# Status fg_get_lin(Vertex *out, FlowGrid *G, linidx_t a);
libocn.fg_get_lin.argtypes = [POINTER(Vertex_C), POINTER(FlowGrid_C), linidx_t]
libocn.fg_get_lin.restype = Status

# Status fg_set_lin(FlowGrid *G, Vertex vert, linidx_t a);
libocn.fg_set_lin.argtypes = [POINTER(FlowGrid_C), Vertex_C, linidx_t]
libocn.fg_set_lin.restype = Status

# Status fg_create_empty(CartPairC dims);
libocn.fg_create_empty.argtypes = [CartPair_C]
libocn.fg_create_empty.restype = POINTER(FlowGrid_C)

# FlowGrid *fg_copy(FlowGrid *G);
libocn.fg_copy.argtypes = [POINTER(FlowGrid_C)]
libocn.fg_copy.restype = POINTER(FlowGrid_C)

# Status fg_destroy(FlowGrid *G);
libocn.fg_destroy.argtypes = [POINTER(FlowGrid_C)]
libocn.fg_destroy.restype = Status

# Status fg_find_upstream_neighbors(linidx_t upstream_indices[8], linidx_t *nupstream, FlowGrid *G, linidx_t a);
libocn.fg_find_upstream_neighbors.argtypes = [linidx_t * 8, POINTER(linidx_t), POINTER(FlowGrid_C), linidx_t]
libocn.fg_find_upstream_neighbors.restype = Status

# Status fg_dfs_iterative(linidx_t upstream_indices[], linidx_t *nupstream, linidx_t idx_stack[], FlowGrid *G, linidx_t a);
libocn.fg_dfs_iterative.argtypes = [POINTER(linidx_t), POINTER(linidx_t), POINTER(linidx_t), POINTER(FlowGrid_C), linidx_t]
libocn.fg_dfs_iterative.restype = Status

# Status flow_downstream(linidx_t downstream_indices[], linidx_t *ndownstream, FlowGrid *G, linidx_t a)
libocn.flow_downstream.argtypes = [POINTER(linidx_t), POINTER(linidx_t), POINTER(FlowGrid_C), linidx_t]
libocn.flow_downstream.restype = Status

##############################
#     OCN.H EQUIVALENTS      #
##############################

# double ocn_compute_energy(FlowGrid *G, double gamma);
libocn.ocn_compute_energy.argtypes = [POINTER(FlowGrid_C), c_double]
libocn.ocn_compute_energy.restype = c_double

# Status ocn_single_erosion_event(FlowGrid *G, double gamma, double temperature, bool calculate_full_energy);
libocn.ocn_single_erosion_event.argtypes = [POINTER(FlowGrid_C), c_double, c_double, c_bool]
libocn.ocn_single_erosion_event.restype = Status

# Status ocn_outer_ocn_loop(FlowGrid *G, uint32_t niterations, double gamma, double *annealing_schedule, bool calculate_full_energy);
libocn.ocn_outer_ocn_loop.argtypes = [POINTER(FlowGrid_C), c_uint32, c_double, POINTER(c_double), c_bool]
libocn.ocn_outer_ocn_loop.restype = Status


__all__ = [
    "SUCCESS",
    "EROSION_FAILURE",
    "OOB_ERROR",
    "NULL_POINTER_ERROR",
    "SWAP_WARNING",
    "MALFORMED_GRAPH_WARNING",
    "libocn",
    "CartPair_C",
    "Vertex_C",
    "FlowGrid_C",
    "IS_ROOT",
]


"""
Utility functions for working with OCNs.
"""

from __future__ import annotations
from collections.abc import Generator
from concurrent.futures import ThreadPoolExecutor
from ctypes import byref
from itertools import product
import os
from typing import Any, Literal, Callable, TYPE_CHECKING, Union
from numbers import Number
import networkx as nx
import numpy as np
from tqdm import tqdm
from functools import partial

import PyOCN._libocn_bindings as _bindings
from PyOCN._statushandler import check_status

if TYPE_CHECKING:
    from .ocn import OCN

_allowed_net_types = {"I", "H", "V", "E"}

#TODO: add ability to move root?
def net_type_to_dag(net_type:Literal["I", "H", "V", "E"], dims:tuple, pbar: bool = False) -> nx.DiGraph:
    """Create a predefined OCN initialization network as a NetworkX DiGraph.

    Parameters
    ----------
    net_type : {"I", "H", "V", "E"}
        The type of network to create.
        Descriptions of allowed types:

        - "I":

          ::

              O--O--O--O--O
                    |
              O--O--O--O--O
                    |
              O--O--O--O--O
                    |
              O--O--X--O--O

        - "V":

          ::

              O  O  O  O  O
               \  \ | /  /
              O  O  O  O  O
               \  \ | /  /
              O  O  O  O  O
               \  \ | /  /
              O--O--X--O--O

        - "H":

          ::

              O  O  O  O
              |  |  | /
              O  O  O--O
              |  | /
              O  O--O--O
              | /
              X--O--O--O

        - "E": A network where every node on the edge of the grid is a root. Initial flow moves away from center towards edges.

    dims : tuple
        The network dimensions as ``(rows, cols)``. Both must be positive even integers.
    pbar : bool, default False
        If True, display a progress bar while constructing the graph.

    Returns
    -------
    networkx.DiGraph
        A directed acyclic graph representing a valid initial OCN configuration.

    Raises
    ------
    ValueError
        If ``net_type`` is invalid or ``dims`` are not two positive even integers.

    Note
    ----
    The returned graph assigns each grid cell exactly one node with a ``pos``
    attribute equal to ``(row, col)``.
    """
    rows, cols = dims
    G = nx.DiGraph()
    pbar = tqdm(product(range(rows), range(cols)), total=rows*cols, disable=not pbar, desc="Creating DAG")
    match net_type:
        case "I":
            jroot = cols // 2
            for i, j in pbar:
                n = i*cols + j
                G.add_node(n, pos=(i, j))
                if j < jroot:
                    G.add_edge(n, n+1)
                elif j > jroot:
                    G.add_edge(n, n-1)
                elif i > 0:
                    G.add_edge(n, n - cols)

        case "V":
            jroot = cols // 2
            for i, j in pbar:
                n = i*cols + j
                G.add_node(n, pos=(i, j))
                if i > 0:
                    if j < jroot:
                        G.add_edge(n, n - cols + 1)
                    elif j > jroot:
                        G.add_edge(n, n - cols - 1)
                    else:
                        G.add_edge(n, n - cols)
                else:
                    if j < jroot:
                        G.add_edge(n, n + 1)
                    elif j > jroot:
                        G.add_edge(n, n - 1)
        case "H": # hip roof is like V, but flowing towards a corner.
            for i, j in pbar:
                n = i*cols + j
                G.add_node(n, pos=(i, j))
                if i == j and i > 0:  # main diagonal
                    G.add_edge(n, n - cols - 1)
                elif i > j:
                    G.add_edge(n, n - cols)
                elif j > i:
                    G.add_edge(n, n - 1)
        case "E":  #TODO: implement a better radial pattern
            half_rows = rows / 2
            half_cols = cols / 2
            for i, j in pbar:
                n = i*cols + j
                G.add_node(n, pos=(i, j))
                # ul quadrant: flow up and left
                if i < half_rows and j < half_cols and i > 0 and j > 0:
                    G.add_edge(n, n - cols - 1)
                # ur quadrant: flow up and right
                elif i < half_rows and j >= half_cols and i > 0 and j < cols - 1:
                    G.add_edge(n, n - cols + 1)
                # ll quadrant: flow down and left
                elif i >= half_rows and j < half_cols and j > 0 and i < rows - 1:
                    G.add_edge(n, n + cols - 1)
                # lr quadrant: flow down and right
                elif i >= half_rows and j >= half_cols and i < rows - 1 and j < cols - 1:
                    G.add_edge(n, n + cols + 1)
        case _:
            raise ValueError(f"Invalid net_type {net_type}. Must be one of {_allowed_net_types}.")
        
    return G

def simulated_annealing_schedule(dims: tuple[int, int],E0: float,constant_phase: float,n_iterations: int,cooling_rate: float,) -> Callable[[int], Union[float, np.ndarray]]:
    """
    Create a simulated-annealing cooling schedule for OCN optimization.

    This returns a callable ``schedule(i)`` that returns the temperature at
    iteration ``i``. The schedule consists of a constant-temperature phase
    followed by an exponentially decaying phase.

    Parameters
    ----------
    dims : tuple[int, int]
        The dimensions of the grid as (rows, cols).
    E0 : float
        Initial energy value.
    constant_phase : float
        Fraction of iterations (``0 <= fraction <= 1``) during which the
        temperature remains constant at ``Energy[0]``.
    n_iterations : int
        Total number of optimization iterations.
    cooling_rate : float
        Positive decay rate controlling the exponential temperature decrease
        after the constant phase.

    Returns
    -------
    Callable[[int], Union[float, np.ndarray]]
        A function mapping an iteration index ``i`` to a temperature value. If
        vectorized evaluation is used, may return a NumPy array of temperatures.

    Note
    ----
    The exponential phase follows the form

    .. math::

        T_i = E_0 \exp\left(-\\frac{r\cdot(i - n_0)}{N}\\right),

    where ``E0`` is the initial energy, ``n0`` is the number of iterations in
    the constant phase, ``r`` is the cooling rate,and ``N = rows * cols``.
    """
    if (not isinstance(constant_phase, Number) or constant_phase < 0 or constant_phase > 1):
        raise ValueError(f"constant_phase must be a number between 0 and 1. Got {constant_phase}")
    if (not isinstance(n_iterations, int) or n_iterations <= 0):
        raise ValueError(f"n_iterations must be a positive integer. Got {n_iterations}")
    if not isinstance(cooling_rate, Number):
        raise ValueError(f"cooling_rate must be a number. Got {cooling_rate}")
    if not isinstance(E0, Number) or E0 <= 0:
        raise ValueError(f"E0 must be a positive number. Got {E0}")
    if not (isinstance(dims, tuple) and len(dims) == 2 and all(isinstance(d, int) and d > 0 for d in dims)):
        raise ValueError(f"dims must be a tuple of two positive integers. Got {dims}")
    
    n_constant = int(constant_phase * n_iterations)
    nnodes = dims[0] * dims[1]

    term1 = -cooling_rate / nnodes
    term2 = cooling_rate * n_constant / nnodes

    def schedule(i):
        i = np.asarray(i)
        return np.where(i < n_constant, E0, E0 * np.exp(term1*i + term2))

    return schedule

def unwrap_digraph(dag: nx.DiGraph, dims: tuple[int, int]) -> nx.DiGraph:
    """"unwrap" gridcell coordinate attributes in a directed acyclic graph to place connected
    nodes adjacent to each other, removing periodic boundary conditions.

    Parameters
    ----------
    dag : nx.DiGraph
        The input directed acyclic graph with periodic boundary conditions.
        Each node must have a 'pos' attribute indicating its (row, col) position.
    dims : tuple[int, int]
        The dimensions of the grid as (rows, cols). Both must be positive integers.

    Returns
    -------
    nx.DiGraph
        A new directed acyclic graph with unwrapped grid coordinates. May not be
        consistent with a grid structure.

    Raises
    ------
    ValueError
        If any node in the input graph lacks a 'pos' attribute or if the
        dimensions are not positive integers.

    Important
    ---------
    The function assumes that the input graph is a valid DAG and that the
    'pos' attributes are correctly assigned. The output graph will no longer
    span a toroidal topology and will no longer cover a dense grid of nodes.
    """
    new_dag = dag.copy()
    
    for _ in range(2):
        for n in nx.topological_sort(new_dag):

            pos = new_dag.nodes[n]['pos']
            r, c = pos
            succs = list(new_dag.successors(n))
            
            # check for row wrapping
            for s in succs:
                sr, _ = new_dag.nodes[s]['pos']
                dr = sr - r
                # if we detect wrapping, move all ancestors and self
                if dr > 1:  # wrapped downward
                    for anc in nx.ancestors(new_dag, n):
                        anc_r, anc_c = new_dag.nodes[anc]['pos']
                        new_dag.nodes[anc]['pos'] = (anc_r + dims[0], anc_c)
                    new_dag.nodes[n]['pos'] = (r + dims[0], c)
                    break
                elif dr < -1:  # wrapped upward
                    for anc in nx.ancestors(new_dag, n):
                        anc_r, anc_c = new_dag.nodes[anc]['pos']
                        new_dag.nodes[anc]['pos'] = (anc_r - dims[0], anc_c)
                    new_dag.nodes[n]['pos'] = (r - dims[0], c)
                    break
            # check for column wrapping
            for s in succs:
                _, sc = new_dag.nodes[s]['pos']
                dc = sc - c
                if dc > 1:  # wrapped rightward
                    for anc in nx.ancestors(new_dag, n):
                        anc_r, anc_c = new_dag.nodes[anc]['pos']
                        new_dag.nodes[anc]['pos'] = (anc_r, anc_c + dims[1])
                    new_dag.nodes[n]['pos'] = (r, c + dims[1])
                    break
                elif dc < -1:  # wrapped leftward
                    for anc in nx.ancestors(new_dag, n):
                        anc_r, anc_c = new_dag.nodes[anc]['pos']
                        new_dag.nodes[anc]['pos'] = (anc_r, anc_c - dims[1])
                    new_dag.nodes[n]['pos'] = (r, c - dims[1])
                    break
    # Adjust positions to be non-negative
    positions = np.array(list(nx.get_node_attributes(new_dag, 'pos').values()))
    row_off, col_off = positions.min(axis=0)
    for n in new_dag.nodes:
        r, c = new_dag.nodes[n]['pos']
        new_dag.nodes[n]['pos'] = (r - row_off, c - col_off)
    return new_dag

def assign_subwatersheds(dag: nx.DiGraph) -> None:
    """Assign a 'watershed_id' attribute to each node in the DAG. 
    The resulting watershed_ids will be of the highest order possible,
    meaning that ids are assigned based on watersheds that drain directly into the 
    root nodes of the graph. To assign ids to lower order watersheds, consider first
    partitioning the graph into smaller subgraphs using the `get_subwatersheds` function.

    Parameters
    ----------
    dag : nx.DiGraph
        The input directed acyclic graph. Each node must have a 'pos' attribute
        indicating its (row, col) position.

    Returns
    -------
    None
        The function modifies the input graph in place by adding a 'watershed_id'
        attribute to each node.

    Raises
    ------
    ValueError
        If any node in the input graph lacks a 'pos' attribute.

    Note
    ----
    A subwatershed is defined as the set of nodes that drain to a common outlet,
    where an outlet is a node with out-degree zero. Each subwatershed is assigned
    a unique integer ID, starting from 0. Nodes that are outlets themselves are
    assigned a watershed ID of -1.
    """
    roots = [n for n, d in dag.out_degree() if d==0]
    subwatershed_outlets = [n for root in roots for n in dag.predecessors(root)]
    subwatersheds = list(set(nx.ancestors(dag, outlet)) | {outlet} for outlet in subwatershed_outlets)
    subwatersheds = [dag.subgraph(wshd) for wshd in subwatersheds]
    for i, wshd in enumerate(subwatersheds):
        nx.set_node_attributes(wshd, i + 1, 'watershed_id')
    for r in roots:
        nx.set_node_attributes(dag, {r: -1}, 'watershed_id')

def get_subwatersheds(dag : nx.DiGraph, node : Any) -> set[nx.DiGraph]:
    """Extract subwatershed subgraphs from the main DAG. Each subwatershed drains to a common outlet node `node`.
    Node `node` is not included in the returned subwatershed graphs.

    Parameters
    ----------
    dag : nx.DiGraph
        The input directed acyclic graph. Each node must have a 'pos' attribute
        indicating its (row, col) position.
    node : Any
        A node in the graph representing the outlet of a subwatershed.

    Returns
    -------
    set of nx.DiGraph
        A set of directed acyclic graphs, each representing a subwatershed.

    Danger
    ------
    The returned subwatersheds are subgraph views of the input graph and share node
    and edge data with the original graph. Unless copied, any changes to node or edge attributes
    in the subwatersheds will affect the original graph.
    """
    subwatershed_outlets = [n for n in dag.predecessors(node)]
    subwatersheds = (nx.ancestors(dag, outlet) | {outlet} for outlet in subwatershed_outlets)
    subwatersheds = set(dag.subgraph(wshd) for wshd in subwatersheds)
    return subwatersheds

# # deferring stream order calculations for now.
# def assign_strahler_orders(dag: nx.DiGraph) -> None:
#     """Assign Strahler order to each node in the DAG as a 'strahler_order' attribute.
#     Modifies the input graph in place.

#     Parameters
#     ----------
#     dag : nx.DiGraph
#         The input directed acyclic graph.
#     """
#     # Initialize Strahler order for leaf nodes
#     leaf_nodes = [n for n, d in dag.in_degree() if d == 0]
#     nx.set_node_attributes(dag, {n: 1 for n in leaf_nodes}, 'strahler_order')

#     # Compute Strahler order for other nodes
#     strahler_orders = dict(nx.get_node_attributes(dag, 'strahler_order'))
#     for n in nx.topological_sort(dag):
#         if dag.in_degree(n):
#             in_orders = [strahler_orders[p] for p in dag.predecessors(n)]
#             max_order = max(in_orders)
#             if sum(o == max_order for o in in_orders) > 1:
#                 order = max_order + 1
#             else:
#                 order = max_order
#             strahler_orders[n] = order
#     nx.set_node_attributes(dag, strahler_orders, 'strahler_order')

# def assign_shreve_orders(dag: nx.DiGraph) -> None:
#     """Assign Shreve order to each node in the DAG as a 'shreve_order' attribute.
#     Shreve order is defined as the total number of upstream sources that contribute to a node.

#     Parameters
#     ----------
#     dag : nx.DiGraph
#         The input directed acyclic graph.
#     """
#     # Initialize Shreve order for leaf nodes
#     leaf_nodes = [n for n, d in dag.in_degree() if d == 0]
#     nx.set_node_attributes(dag, {n: 1 for n in leaf_nodes}, 'shreve_order')

#     # Compute Shreve order for other nodes
#     shreve_orders = dict(nx.get_node_attributes(dag, 'shreve_order'))
#     for n in nx.topological_sort(dag):
#         if dag.in_degree(n):
#             in_orders = [shreve_orders[p] for p in dag.predecessors(n)]
#             order = sum(in_orders)
#             shreve_orders[n] = order
#     nx.set_node_attributes(dag, shreve_orders, 'shreve_order')
    

def parallel_fit(
    ocn:OCN|list[OCN], 
    n_runs=5, 
    n_threads=None, 
    fit_method:Literal["fit", "fit_custom_cooling"]|list[Literal["fit", "fit_custom_cooling"]]|None=None, 
    increment_rng:bool=True, 
    pbar:bool=False, 
    fit_kwargs:dict|list[dict]=None
) -> tuple[list[Any], list[OCN]]:
    """Convenience function to perform multiple OCN fitting operations in parallel using multithreading. 
    Useful for doing sensitivity analysis or ensemble fitting.

    Parameters
    ----------
    ocn : OCN | list[OCN]
        The OCN instance(s) to fit. If a list of OCNs is provided, each OCN will be fitted independently.
        If a single OCN is provided, it will be copied `n_runs` times.
        Not modified during fitting.
    n_runs : int, default 5
        The number of fitting runs to perform. Must be equal to the length of `ocn` if `ocn` is a list.
    n_threads : int, default None
        The number of worker threads to use for parallel execution. If None, defaults to the number of CPU cores, times 2.
    fit_method : {"fit", "fit_custom_cooling"} | list[{"fit", "fit_custom_cooling"} | None], default None
        The fitting method to use. If None, defaults to "fit". If a list is provided, it must be of length `n_runs`,
        and each element specifies the fitting method for the corresponding fit.
    increment_rng : bool, default True
        If True, each fit's random seed is set to `ocn.rng + i`, where `i` is the thread index.
    pbar : bool, default False
        If True, display a master progress bar that tracks the completion of each thread.
    fit_kwargs : dict | list[dict], optional
        Additional keyword arguments to pass to the `OCN.fit` method. If a list of dicts is provided,
        it must be of length `n_runs`, and each dict will be used for the corresponding fit.

    Returns
    -------
    tuple[list[Any], list[OCN]]
        A tuple containing two lists:
        - A list of results from each fitting operation.
        - A list of fitted OCN instances corresponding to each fitting operation.
    """

    if isinstance(ocn, list):
        if len(ocn) != n_runs:
            raise ValueError(f"When ocn is a list, its length must equal n_runs. Got len(ocn)={len(ocn)} and n_runs={n_runs}.")
        ocn = [o.copy() for o in ocn]
    else:
        ocn = [ocn.copy() for _ in range(n_runs)]

    if isinstance(fit_kwargs, list):
        if len(fit_kwargs) != n_runs:
            raise ValueError(f"When fit_kwargs is a list, its length must equal n_runs. Got len(fit_kwargs)={len(fit_kwargs)} and n_runs={n_runs}.")
        if not all(isinstance(k, dict) for k in fit_kwargs):
            raise ValueError("All elements of fit_kwargs list must be dictionaries.")
    elif not isinstance(fit_kwargs, dict) and fit_kwargs is not None:
        raise ValueError(f"fit_kwargs must be a dict or a list of dicts. Got {type(fit_kwargs)}.")
    else: 
        fit_kwargs = [fit_kwargs or dict() for _ in range(n_runs)]

    if isinstance(fit_method, list):
        if len(fit_method) != n_runs:
            raise ValueError(f"When fit_method is a list, its length must equal n_runs. Got len(fit_method)={len(fit_method)} and n_runs={n_runs}.")
        elif not all(m is None or m in {"fit", "fit_custom_cooling"} for m in fit_method):
            raise ValueError("All elements of fit_method list must be one of 'fit', 'fit_custom_cooling', or None.")
    else:
        if fit_method is not None and fit_method not in {"fit", "fit_custom_cooling"}:
            raise ValueError(f"Invalid fit_method {fit_method}. Must be one of 'fit', 'fit_custom_cooling', or None.")
        fit_method = [fit_method for _ in range(n_runs)]

    def fit(ocn, i, fit_method, fit_kwargs):
        if increment_rng:
            ocn.rng  = ocn.rng + i
        if fit_method is None or fit_method == "fit":
            res = ocn.fit(**fit_kwargs)
        elif fit_method == "fit_custom_cooling":
            res = ocn.fit_custom_cooling(**fit_kwargs)
        else:
            raise ValueError(f"Invalid fit_method {fit_method}. Must be one of 'fit', 'fit_custom_cooling', or None.")
        return res, i  # return index to place result correctly, in case of out-of-order completion.
    
    if n_threads is None:
        n_threads = os.cpu_count()*2
    n_threads = min(os.cpu_count()*2, n_threads)

    with ThreadPoolExecutor(max_workers=n_threads) as executor:
        futures = []
        for i in range(n_runs):
            futures.append(executor.submit(fit, ocn[i], i, fit_method[i], fit_kwargs[i]))
        results = [None] * n_runs
        pbar = tqdm(futures, disable=not pbar, desc="Fitting OCNs")
        for future in futures:
            res, idx = future.result()
            results[idx] = res
            pbar.update(1)
    return results, ocn

# # Commenting out for now. May decide to re-introduce later
# def ancestors(ocn:OCN, pos:tuple[int, int]) -> set[tuple[int, int]]:
#     """Returns all nodes that drain into the node at position `pos` in the OCN.

#     Parameters
#     ----------
#     ocn : OCN
#         The OCN instance.
#     pos : tuple[int, int]
#         The (row, col) position of the node whose predecessors are to be found.

#     Returns
#     -------
#     set()
#         A set of (row, col) tuples representing the positions of all ancestor nodes
#         that eventually drain into the specified node, not including the node itself.

#     See Also
#     --------
#     :meth:`OCN.predecessors`
#     :meth:`OCN.successors`
#     :meth:`utils.descendants`
#     """

#     pos = tuple(pos)
#     if len(pos) != 2 or not all(isinstance(p, int) for p in pos):
#         raise TypeError(f"Position must be a tuple of two integers. Got {pos}.")
#     if (pos[0] < 0 or pos[0] >= ocn.dims[0]) or (pos[1] < 0 or pos[1] >= ocn.dims[1]):
#         raise IndexError(f"Position {pos} is out of bounds for OCN with dimensions {ocn.dims}.")

#     a = _bindings.libocn.fg_cart_to_lin(
#         _bindings.CartPair_C(row=pos[0], col=pos[1]),
#         _bindings.CartPair_C(row=ocn.dims[0], col=ocn.dims[1])
#     )
#     upstream_indices = (_bindings.linidx_t * (ocn.dims[0]*ocn.dims[1]))()
#     nupstream = _bindings.linidx_t(0)
#     check_status(_bindings.libocn.fg_dfs_iterative(
#         upstream_indices, 
#         byref(nupstream), 
#         (_bindings.linidx_t * (ocn.dims[0]*ocn.dims[1]))(), 
#         ocn._OCN__p_c_graph, 
#         a
#     ))

#     return set(
#         (int(c.row), int(c.col)) 
#         for c in (
#             _bindings.libocn.fg_lin_to_cart(
#                 idx, 
#                 _bindings.CartPair_C(row=ocn.dims[0], col=ocn.dims[1])
#             ) 
#             for idx in upstream_indices[:nupstream.value]
#         )
#     )

# # TODO write tests for the traversal functions
# def descendants(ocn:OCN, pos:tuple[int, int]) -> set[tuple[int, int]]:
#     """Returns all nodes that are reachable from the node at position `pos` in the OCN.
#     Mirrors the functionality of `networkx.descendants`.

#     Parameters
#     ----------
#     ocn : OCN
#         The OCN instance.
#     pos : tuple[int, int]
#         The (row, col) position of the node whose successors are to be found.

#     Returns
#     -------
#     set()
#         A set of (row, col) tuples representing the positions of all descendant nodes
#         that the specified node eventually drains into, not including the node itself.

#     See Also
#     --------
#     :meth:`OCN.predecessors`
#     :meth:`OCN.successors`
#     :meth:`utils.ancestors`
#     """

#     pos = tuple(pos)
#     if len(pos) != 2 or not all(isinstance(p, int) for p in pos):
#         raise TypeError(f"Position must be a tuple of two integers. Got {pos}.")
#     if (pos[0] < 0 or pos[0] >= ocn.dims[0]) or (pos[1] < 0 or pos[1] >= ocn.dims[1]):
#         raise IndexError(f"Position {pos} is out of bounds for OCN with dimensions {ocn.dims}.")

#     a = _bindings.libocn.fg_cart_to_lin(
#         _bindings.CartPair_C(row=pos[0], col=pos[1]),
#         _bindings.CartPair_C(row=ocn.dims[0], col=ocn.dims[1])
#     )
#     downstream_indices = (_bindings.linidx_t * (ocn.dims[0]*ocn.dims[1]))()
#     ndownstream = _bindings.linidx_t(0)
#     check_status(_bindings.libocn.flow_downstream(
#         downstream_indices,
#         byref(ndownstream),
#         ocn._OCN__p_c_graph,
#         a
#     ))

#     return set(
#         (int(c.row), int(c.col))
#         for c in (
#             _bindings.libocn.fg_lin_to_cart(
#                 idx,
#                 _bindings.CartPair_C(row=ocn.dims[0], col=ocn.dims[1])
#             )
#             for idx in downstream_indices[:ndownstream.value]
#         )
#     )

__all__ = [
    "net_type_to_dag",
    "simulated_annealing_schedule",
    "unwrap_digraph",
    "assign_subwatersheds",
    "get_subwatersheds",
    "parallel_fit",
    "assign_strahler_orders",
    "assign_shreve_orders",
]


"""
Basic test suite for PyOCN.

These tests verify core functionality with deterministic outputs.
"""
import unittest
from time import perf_counter as timer
import numpy as np
import networkx as nx
import PyOCN as po

class TestBasicOCN(unittest.TestCase):
    """Basic tests for OCN creation and energy computation."""
    
    def test_ocn_to_xarray(self):
        """Test that OCN can be converted to xarray and back."""
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(32, 32),
            random_state=1234,
        )
        ocn.to_xarray()
        
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(32, 32),
            random_state=1234,
            wrap=True,
        )
        ocn.to_xarray()
        
    def test_ocn_energy(self):
        """Test that OCN creation produces expected energy value."""
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(64, 64),
            random_state=8472,
        )
        
        expected_energy = 16469.684
        actual_energy = ocn.energy
        
        self.assertAlmostEqual(
            actual_energy, 
            expected_energy, 
            places=3,
            msg=f"Expected energy {expected_energy}, got {actual_energy}"
        )
    
    def test_ocn_copy(self):
        """Test that copying an OCN instance produces an identical copy."""
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(32, 32),
            random_state=1234,
        )
        
        ocn_copy = ocn.copy()
        
        # Check that the copy has the same attributes
        self.assertEqual(ocn.dims, ocn_copy.dims, "Dimensions do not match.")
        self.assertEqual(ocn.energy, ocn_copy.energy, "Energies do not match.")
        self.assertEqual(ocn.nroots, ocn_copy.nroots, "Number of roots do not match.")
        self.assertEqual(ocn.wrap, ocn_copy.wrap, "Wrap settings do not match.")
        self.assertEqual(ocn.gamma, ocn_copy.gamma, "Gamma values do not match.")
        self.assertEqual(ocn.vertical_exaggeration, ocn_copy.vertical_exaggeration, "Vertical exaggeration values do not match.")
        
        # Check that the internal data arrays are equal
        original_array = ocn.to_numpy()
        copy_array = ocn_copy.to_numpy()
        
        np.testing.assert_array_equal(
            original_array, 
            copy_array, 
            err_msg="Internal data arrays do not match."
        )

        # fit both and check energies are still equal
        ocn.fit(array_reports=0)
        ocn_copy.fit(array_reports=0)
        
        self.assertEqual(ocn.energy, ocn_copy.energy, "Energies do not match after fitting.")

    def test_ocn_custom_cooling(self):
        """Test that custom cooling schedule affects energy as expected."""
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(64, 64),
            random_state=143798,
        )
        
        # Fit with a custom cooling schedule
        energy = ocn.energy
        def custom_schedule(iter):
            return energy / (iter + 1)
        
        ocn_copy = ocn.copy()

        ocn_copy.fit_custom_cooling(custom_schedule, max_iterations_per_loop=100, n_iterations=50_000, array_reports=0, iteration_start=0)
        ocn.fit_custom_cooling(custom_schedule, max_iterations_per_loop=100, n_iterations=5_000, array_reports=0, iteration_start=0)
        ocn.fit_custom_cooling(custom_schedule, max_iterations_per_loop=100, n_iterations=45_000, array_reports=0, iteration_start=5_000)

        self.assertAlmostEqual(ocn.energy, ocn_copy.energy, places=6)

    def test_ocn_rng(self):
        """Test that the rng property returns the expected random state."""
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(64, 64),
            random_state=8472,
        )
        expected_rng = 40451845669725511932870495160880127104
        
        self.assertEqual(
            ocn.rng, 
            expected_rng, 
            f"Expected RNG state {expected_rng}, got {ocn.rng}"
        )

        ocn.rng = 8470
        self.assertNotEqual(
            ocn.rng, 
            expected_rng, 
            f"After changing seed, RNG state should differ from {expected_rng}"
        )

        # Reset to original seed and check
        ocn.rng = 8472
        self.assertEqual(
            ocn.rng, 
            expected_rng, 
            f"After resetting, expected RNG state {expected_rng}, got {ocn.rng}"
        )

        ocn.rng = 8472
        ocn.rng = 8472
        ocn.rng = 8472
        self.assertEqual(
            ocn.rng, 
            expected_rng, 
            f"Setting rng to the same seed multiple times should not change the state. Expected {expected_rng}, got {ocn.rng}"
        )

        ocn.single_iteration(0)
        self.assertNotEqual(
            ocn.rng, 
            expected_rng, 
            f"After iteration, RNG state should differ from {expected_rng}. Got {ocn.rng}"
        )

    def test_from_digraph(self):
        """Test creating OCN from custom NetworkX digraph."""
        # Create a simple 3x3 cross pattern like in demo
        dag = nx.DiGraph()
        for i in range(9):
            row, col = divmod(i, 3)
            dag.add_node(i, pos=(row, col))
        
        # Simple flow pattern: all flow to center (node 4)
        for i in [0, 1, 2, 3, 5, 6, 7, 8]:
            dag.add_edge(i, 4)
        
        ocn = po.OCN.from_digraph(dag, random_state=1234)
        
        self.assertEqual(ocn.dims, (3, 3))
        self.assertEqual(ocn.nroots, 1)  # Only center node should be root
        self.assertAlmostEqual(ocn.energy, 11.0, places=6)
        
        ocn = po.OCN.from_digraph(dag, random_state=1234, gamma=1)
        self.assertAlmostEqual(ocn.energy, 17.0, places=6)

    def test_export_formats(self):
        """Test different export formats produce consistent data."""
        ocn = po.OCN.from_net_type("V", dims=(10, 10), random_state=9999)
        
        # Test numpy export
        numpy_array = ocn.to_numpy(unwrap=False)
        self.assertEqual(numpy_array.shape, (3, 10, 10))  # 3 channels: energy, area, watershed
        
        # Test digraph export  
        dag = ocn.to_digraph()
        self.assertEqual(len(dag.nodes), 100)  # 10x10 = 100 nodes
        self.assertTrue(nx.is_directed_acyclic_graph(dag))
        
        # Check that node attributes exist
        for node in dag.nodes:
            attrs = dag.nodes[node]
            self.assertIn('pos', attrs)
            self.assertIn('drained_area', attrs)
            self.assertIn('energy', attrs)
            self.assertIn('elevation', attrs)

    def test_periodic_boundaries(self):
        """Test OCN with periodic boundary conditions."""
        ocn_wrap = po.OCN.from_net_type("H", dims=(10, 10), wrap=True, random_state=5555)
        ocn_no_wrap = po.OCN.from_net_type("H", dims=(10, 10), wrap=False, random_state=5555)
        ocn_wrap.fit(n_iterations=500, pbar=False)
        ocn_no_wrap.fit(n_iterations=500, pbar=False)
        
        self.assertTrue(ocn_wrap.wrap)
        self.assertFalse(ocn_no_wrap.wrap)
        
        # Both should have same dimensions but potentially different energies
        self.assertEqual(ocn_wrap.dims, ocn_no_wrap.dims)
        # Wrapped version typically has different energy due to edge connections
        self.assertNotEqual(ocn_wrap.energy, ocn_no_wrap.energy)

    def test_fit_convergence(self):
        """Test that early exit works as intended."""
        ocn = po.OCN.from_net_type("E", dims=(20, 20), random_state=7777)
        
        ocn.fit(n_iterations=20*20*500, pbar=False, tol=1e-4)
        final_energy = ocn.energy

        self.assertIsInstance(final_energy, float)
        self.assertGreater(len(ocn.history), 0)
        
        # History should have correct structure: [iteration, energy, temperature]
        self.assertEqual(ocn.history.shape[1], 3)
        self.assertTrue(np.all(ocn.history[:, 0] >= 0))  # iterations >= 0
        self.assertTrue(np.all(ocn.history[:, 1] > 0))   # energy > 0
        self.assertTrue(np.all(ocn.history[:, 2] >= 0))  # temperature >= 0
        
        # final energy should match last history entry
        self.assertEqual(ocn.energy, ocn.history[-1, 1])

        # iteration should have stopped before max iterations and should be less than tol
        self.assertLess(ocn.history[-1, 0], 20*20*500)
        self.assertLessEqual((ocn.history[-1, 1] - ocn.history[-2, 1])/ocn.history[-1, 1], 1e-4)

    def test_single_iteration(self):
        """Test single iteration method."""
        ocn = po.OCN.from_net_type("V", dims=(16, 16), random_state=3333)
        initial_history_len = len(ocn.history)
        
        result = ocn.single_iteration(temperature=ocn.energy, array_report=False)
        
        # Should return None when array_report=False
        self.assertIsNone(result)
        
        # History should have one more entry
        self.assertEqual(len(ocn.history), initial_history_len + 1)

    def test_error_handling(self):
        """Test error handling for invalid inputs."""
        # Invalid gamma
        with self.assertRaises(TypeError):
            po.OCN.from_net_type("V", dims=(10, 10), gamma="invalid")
        
        # Invalid net_type
        with self.assertRaises(ValueError):
            po.OCN.from_net_type("INVALID", dims=(10, 10))
        
        # Invalid dimensions
        with self.assertRaises((ValueError, TypeError)):
            po.OCN.from_net_type("V", dims=(5,))  # Should need 2D dims

    def test_greedy_optimization(self):
        ocn = po.OCN.from_net_type(
            net_type="V",
            dims=(16, 16),
            wrap=True,
            random_state=8472,
        )
        # use a temperature of 1e-7 since we were bitten before by accidentally using 1/temperature. 
        # Using temperature=0 could lead to infinities, which might behave unexpectedly.
        # this way, if our mass is wrong and blows up somewhere, we know that we will catch it.
        ocn.fit_custom_cooling(lambda t: np.ones_like(t)*1e-7, pbar=False, n_iterations=16**2*100, max_iterations_per_loop=1)
        self.assertLessEqual(np.quantile(np.diff(ocn.history[:, 1]), 0.999) - 1e-7, 0, "Energy did not decrease monotonically.")
        
        ocn = po.OCN.from_net_type(
            net_type="E",
            dims=(16, 16),
            wrap=True,
            random_state=8472,
        )
        ocn.fit_custom_cooling(lambda t: np.ones_like(t)*1e-7, pbar=False, n_iterations=16**2*100, max_iterations_per_loop=1)
        self.assertLessEqual(np.quantile(np.diff(ocn.history[:, 1]), 0.999) - 1e-7, 0, "Energy did not decrease monotonically.")

    def test_parallel_fit(self):
        """Test parallel fitting utility."""

        ocn= po.OCN.from_net_type(net_type="I", dims=(42, 40), random_state=542390)
        t0 = timer()
        ocn.fit()
        t1 = timer()
        single_run_time = t1 - t0
        energy = ocn.energy

        ocn = po.OCN.from_net_type(net_type="I", dims=(42, 40), random_state=542390)
        t0 = timer()
        _, fitted_ocns = po.utils.parallel_fit(
            ocn=ocn,
            n_runs=5,
            increment_rng=False,
        )
        t1 = timer()
        parallel_time = t1 - t0
        energies = [o.energy for o in fitted_ocns]
        
        self.assertEqual(len(fitted_ocns), 5, "Number of fitted OCNs does not match number of runs.")

        for e in energies:
            self.assertAlmostEqual(e, energy, places=6, msg="Energies from parallel fits do not match.")
        print()
        print(f"\tSingle run time: {single_run_time:.2f} seconds")
        print(f"\tParallel fit time for 5 runs: {parallel_time:.2f} seconds")
        self.assertLess(parallel_time, single_run_time * 4, "Parallel fitting did not speed up the process as expected.")

        for o in fitted_ocns:
            self.assertAlmostEqual(ocn.energy, o.history[0, 1], places=6, msg="Initial energy was not preserved.")

    def test_energy_update_method(self):
        ocn = po.OCN.from_net_type("E", dims=(44, 44), random_state=238155)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=True)
        energy = ocn.energy
        ocn = po.OCN.from_net_type("E", dims=(44, 44), random_state=238155)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=False)
        energy_2 = ocn.energy
        self.assertAlmostEqual(energy/energy_2, 1.0, places=3, msg="Energies do not match between full_energy_calc and incremental update methods.")
        
        ocn = po.OCN.from_net_type("E", dims=(31, 31), random_state=12638, gamma=0.21)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=True)
        energy = ocn.energy
        ocn = po.OCN.from_net_type("E", dims=(31, 31), random_state=12638, gamma=0.21)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=False)
        energy_2 = ocn.energy
        self.assertAlmostEqual(energy/energy_2, 1.0, places=3, msg="Energies do not match between full_energy_calc and incremental update methods.")
        
        ocn = po.OCN.from_net_type("I", dims=(38, 38), random_state=19075, gamma=0.78)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=True)
        energy = ocn.energy
        ocn = po.OCN.from_net_type("I", dims=(38, 38), random_state=19075, gamma=0.78)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=False)
        energy_2 = ocn.energy
        self.assertAlmostEqual(energy/energy_2, 1.0, places=3, msg="Energies do not match between full_energy_calc and incremental update methods.")
        
        ocn = po.OCN.from_net_type("I", dims=(20, 41), random_state=910536)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=True)
        energy = ocn.energy
        ocn = po.OCN.from_net_type("I", dims=(20, 41), random_state=910536)
        ocn.fit(pbar=False, max_iterations_per_loop=10_000, calculate_full_energy=False)
        energy_2 = ocn.energy

        self.assertAlmostEqual(energy/energy_2, 1.0, places=3, msg="Energies do not match between full_energy_calc and incremental update methods.") 

    def test_gamma(self):
        """Test that different gamma values affect energy as expected."""
        ocn_gamma_1 = po.OCN.from_net_type(
            net_type="V",
            dims=(32, 32),
            gamma=1.0,
            random_state=1234,
        )
        energy_gamma_1 = ocn_gamma_1.energy
        
        ocn_gamma_2 = po.OCN.from_net_type(
            net_type="V",
            dims=(32, 32),
            gamma=0.5,
            random_state=1234,
        )
        energy_gamma_2 = ocn_gamma_2.energy
        
        ocn_gamma_3 = po.OCN.from_net_type(
            net_type="V",
            dims=(32, 32),
            gamma=0.25,
            random_state=1234,
        )
        energy_gamma_3 = ocn_gamma_3.energy
        
        self.assertNotAlmostEqual(energy_gamma_1, energy_gamma_2, places=1, msg="Energies for gamma=1.0 and gamma=0.5 should differ.")
        self.assertNotAlmostEqual(energy_gamma_1, energy_gamma_3, places=1, msg="Energies for gamma=1.0 and gamma=0.25 should differ.")
        self.assertNotAlmostEqual(energy_gamma_2, energy_gamma_3, places=1, msg="Energies for gamma=0.5 and gamma=0.25 should differ.")
        self.assertLess(energy_gamma_3, energy_gamma_2, msg="Energy should decrease with lower gamma.")
        self.assertLess(energy_gamma_2, energy_gamma_1, msg="Energy should decrease with lower gamma.")
        
        ocn_gamma_1 = po.OCN.from_net_type(
            net_type="E",
            dims=(32, 32),
            gamma=1.0,
            random_state=1234,
        )
        energy_gamma_1 = ocn_gamma_1.energy
        
        ocn_gamma_2 = po.OCN.from_net_type(
            net_type="E",
            dims=(32, 32),
            gamma=0.5,
            random_state=1234,
        )
        energy_gamma_2 = ocn_gamma_2.energy
        
        ocn_gamma_3 = po.OCN.from_net_type(
            net_type="E",
            dims=(32, 32),
            gamma=0.25,
            random_state=1234,
        )
        energy_gamma_3 = ocn_gamma_3.energy
        
        self.assertNotAlmostEqual(energy_gamma_1, energy_gamma_2, places=1, msg="Energies for gamma=1.0 and gamma=0.5 should differ.")
        self.assertNotAlmostEqual(energy_gamma_1, energy_gamma_3, places=1, msg="Energies for gamma=1.0 and gamma=0.25 should differ.")
        self.assertNotAlmostEqual(energy_gamma_2, energy_gamma_3, places=1, msg="Energies for gamma=0.5 and gamma=0.25 should differ.")
        
        self.assertLess(energy_gamma_3, energy_gamma_2, msg="Energy should decrease with lower gamma.")
        self.assertLess(energy_gamma_2, energy_gamma_1, msg="Energy should decrease with lower gamma.")

    def test_watershed_partitioning(self):
        ocn = po.OCN.from_net_type("E", dims=(32, 32), random_state=83)
        ocn.fit(pbar=True)
        G = ocn.to_digraph()
        n = 997
        subgraphs = po.utils.get_subwatersheds(G, n)

        node_check = [
            set([932, 931, 964]),
            set([965]),
            set([900, 901, 902, 903, 904, 933, 934, 935, 936, 837, 838, 967, 840, 839, 966, 868, 869, 870, 871, 872])
        ]
        for wshd in subgraphs:
            self.assertIn(set(wshd.nodes), node_check)

if __name__ == "__main__":
    unittest.main()


from setuptools import setup, Extension, find_packages
import sys
from pathlib import Path


build_dir = Path("PyOCN")

c_src = build_dir / "c_src"
include_dirs = [str(c_src)]

sources = [
    c_src / "ocn.c",
    c_src / "flowgrid.c",
    c_src / "status.c",
    c_src / "rng.c",
    c_src / "pyinit.c",
]
sources = [str(s) for s in sources]

if sys.platform.startswith(("linux", "darwin")):
    extra_compile_args = ["-O3", "-flto", "-fPIC", "-std=c99", "-Wall", "-pedantic", "-march=native"]
    extra_link_args = ["-O3", "-flto"]
    # Link libm for pow() on Unix
    libraries = ["m"]
elif sys.platform.startswith("win"):
    extra_compile_args = ["/O2"]
    extra_link_args = []
    libraries = []
else:
    extra_compile_args = []
    extra_link_args = []
    libraries = []

# Define libocn as an extension module
# makes it possible to do `from PyOCN import libocn` if
# we ever decide to push expose the C API directly to the
# python runtime, instead of treating it as a foreign library.
ext = Extension(
    name="PyOCN._libocn",
    sources=sources,
    include_dirs=include_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    libraries=libraries,
)

setup(
    packages=find_packages(include=["PyOCN", "PyOCN.*"]),
    ext_modules=[ext],
    package_data={"PyOCN": []},  # this is where non-code or data files might go.
)


"""
Plotting utilities for PyOCN.

This module provides convenience functions to visualize OCNs using Matplotlib
and NetworkX.

Note
----
- Node positions are stored as ``pos=(row, col)`` in the data model. For
    visualization, these are converted to cartesian coordinates so that 
    (row, col) = (0, 0) is at the bottom-left corner.
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING, Literal

import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
import warnings

if TYPE_CHECKING:
    from .ocn import OCN
from .utils import unwrap_digraph

def _pos_to_xy(dag: nx.DiGraph, nrows=None) -> dict[Any, tuple[float, float]]:
    """
    Convert node ``pos`` from (row, col) to plotting coordinates (x, y).

    Parameters
    ----------
    dag : nx.DiGraph
        Graph whose nodes have ``pos=(row, col)`` attributes.
    nrows : int, optional
        Number of rows in the grid. If ``None``, it is inferred from the
        maximum row index in the node positions.

    Returns
    -------
    dict[Any, tuple[float, float]]
        Mapping from node to ``(x, y)`` where ``x=col`` and ``y`` is flipped
        to plot with origin at the bottom.

    Note
    ----
    The vertical coordinate is transformed as ``y = nrows - row - 1`` to
    match typical plotting conventions.
    """
    pos = nx.get_node_attributes(dag, 'pos')
    nrows = nrows if nrows is not None else max(r for r, _ in pos.values()) + 1
    for node, (r, c) in pos.items():
        pos[node] = (c, nrows - r - 1)  
    return pos

def plot_ocn_as_dag(ocn: OCN, attribute: str | None = None, ax=None, norm=None, **kwargs):
    """
    Plot the OCN as a DAG using NetworkX.

    Parameters
    ----------
    ocn : OCN
        The OCN instance to plot.
    attribute : str, optional
        Node attribute name for coloring (e.g., ``'drained_area'`` or
        ``'energy'``). If omitted, a uniform color is used.
    ax : matplotlib.axes.Axes, optional
        Target axes. If ``None``, a new figure and axes are created.
    norm : matplotlib.colors.Normalize, optional
        Normalization to apply to node colors when ``attribute`` is provided.
        If specified, any ``vmin``/``vmax`` passed in ``**kwargs`` are ignored.
    **kwargs
        Additional keyword arguments forwarded to
        :func:`networkx.draw_networkx` (e.g., ``cmap``, ``vmin``, ``vmax``,
        size and style options).

    Returns
    -------
    tuple
        A pair ``(artists, ax)`` where ``artists`` is the object returned by
        ``networkx.draw_networkx`` (backend-dependent; often ``None``) and
        ``ax`` is the axes used for drawing.
    """
    
    dag = ocn.to_digraph()
    if ocn.wrap:
        dag = unwrap_digraph(dag, ocn.dims)
    pos = _pos_to_xy(dag, nrows=ocn.dims[0])

    if ax is None:
        _, ax = plt.subplots()

    node_color = "C0"
    if attribute is not None:
        node_color = list(nx.get_node_attributes(dag, attribute).values())

    if norm is not None:
        if ("vmin" in kwargs or "vmax" in kwargs):
            warnings.warn("norm is specified, ignoring vmin/vmax.")
        kwargs["vmin"] = 0
        kwargs["vmax"] = 1
        node_color = norm(node_color)
        
    p = nx.draw_networkx(dag, node_color=node_color, pos=pos, ax=ax, **kwargs)
    return p, ax

def plot_ocn_raster(ocn: OCN, attribute:Literal['energy', 'drained_area', 'elevation']='elevation', ax=None, **kwargs):
    """
    Plot a raster image of grid cell energies.

    Parameters
    ----------
    ocn : OCN
        The OCN instance to plot.
    attribute : str, default 'elevation'
        The node attribute to visualize ('energy', 'drained_area', 'elevation').
    ax : matplotlib.axes.Axes, optional
        Target axes. If ``None``, a new figure and axes are created.
    **kwargs
        Additional keyword arguments forwarded to ``imshow`` (e.g.,
        ``cmap``, interpolation options, and normalization).

    Returns
    -------
    matplotlib.axes.Axes
        The axes containing the rendered image.
    matplotlib.image.AxesImage
        The image artist created by ``imshow``.
    """

    array = ocn.to_numpy(unwrap=ocn.wrap)
    match attribute:
        case 'energy':
            array = array[0]
        case 'drained_area':
            array = array[1]
        case 'elevation':
            array = array[2]
        case _:
            raise ValueError(f"Unknown attribute '{attribute}'. Must be one of 'energy', 'drained_area', or 'elevation'.")

    if "cmap" not in kwargs:
        kwargs["cmap"] = "terrain"
        
    if ax is None:
        _, ax = plt.subplots()

    im = ax.imshow(array, **kwargs)
    return ax, im
    

def plot_positional_digraph(dag: nx.DiGraph, ax=None, nrows:int|None=None, **kwargs):
    """
    Plot a DAG with node positions taken from their ``pos`` attributes.

    Parameters
    ----------
    dag : nx.DiGraph
        Graph whose nodes have ``pos=(row, col)``.
    ax : matplotlib.axes.Axes, optional
        Target axes. If ``None``, a new figure and axes are created.
    nrows : int | None, default None
        Number of rows in the grid. If ``None``, infer from the maximum row index in
        the node positions. If ``False``, the vertical coordinates are not
    **kwargs
        Additional keyword arguments forwarded to
        :func:`networkx.draw_networkx`.

    Returns
    -------
    tuple
        A pair ``(artists, ax)`` where ``artists`` is the object returned by
        ``networkx.draw_networkx`` (often ``None``) and ``ax`` is the axes
        used for drawing.
    """
    pos = _pos_to_xy(dag, nrows)

    if ax is None:
        _, ax = plt.subplots()

    p = nx.draw_networkx(dag, pos=pos, ax=ax, **kwargs)
    return p, ax

__all__ = [
    "plot_ocn_as_dag",
    "plot_ocn_raster",
    "plot_positional_digraph",
]


"""
Functions for converting between NetworkX graphs and FlowGrid_C structures.
"""

from ctypes import byref, POINTER
from itertools import product
import warnings

from numpy import integer as np_integer
import networkx as nx

from ._statushandler import check_status
from . import _libocn_bindings as _bindings

def to_digraph(c_graph:_bindings.FlowGrid_C, gamma: float|None = None, vertical_exaggeration: float = 1.0) -> nx.DiGraph:
    """Convert the FlowGrid_C to a NetworkX directed graph.
    
    Parameters
    ----------
    c_graph : _bindings.FlowGrid_C
        The ctypes object representing the underlying flowgrid
    gamma : float | None
        The gamma parameter for fitting the OCN. Needed to compute elevation, slope, and stream length. Default None.
    vertical_exaggeration : float
        How much to exaggerate vertical elevation values. Default 1.0 (no exaggeration)

    Returns
    -------
    nx.DiGraph
        
    """

    dag = nx.DiGraph()
    vert_c = _bindings.Vertex_C()
    pbar = product(range(c_graph.dims.row), range(c_graph.dims.col))
    for r, c in pbar:
        a = _bindings.libocn.fg_cart_to_lin(
            _bindings.CartPair_C(row=r, col=c), 
            c_graph.dims
        )
        check_status(_bindings.libocn.fg_get_lin(
            byref(vert_c), 
            byref(c_graph),
            a,
        ))
        dag.add_node(
            a, 
            pos=(r, c), 
            drained_area=vert_c.drained_area,
            _adown=vert_c.adown,
            _edges=vert_c.edges,
            _downstream=vert_c.downstream,
            _visited=vert_c.visited,
        )
        if vert_c.downstream != _bindings.IS_ROOT:
            dag.add_edge(a, vert_c.adown)

    # compute slope, length, and elevation if gamma is given
    if gamma is not None:
        resolution = c_graph.resolution
        for parent, child in dag.edges:
            slope = vertical_exaggeration*dag.nodes[parent]["drained_area"]**(gamma - 1)
            # length of diagonal vs orthogonal edges
            if dag.nodes[parent]["_downstream"] % 2 == 0:
                length = resolution
            else:
                length = (2**0.5) * resolution
            dag.edges[parent, child]["slope"] = slope
            dag.edges[parent, child]["length"] = length
        for n in list(nx.topological_sort(dag))[::-1]:
            successors = list(dag.successors(n))
            if len(successors) == 0:
                dag.nodes[n]["elevation"] = 0.0
            for i, child in enumerate(successors):
                if i > 1: 
                    raise ValueError("Node has more than one successor, cannot compute elevation.")
                slope, length = dag.edges[n, child]["slope"], dag.edges[n, child]["length"]
                dag.nodes[n]["elevation"] = slope * length + dag.nodes[child]["elevation"]
    return dag

def from_digraph(G: nx.DiGraph, resolution:float=1, verbose:bool=False, validate:bool=True, wrap:bool=False) -> POINTER:
    """
    Convert a NetworkX directed graph into a FlowGrid_C. Called by the OCN constructor.

    Parameters
    ----------
    G: nx.DiGraph
        The digraph object to initialize from
    resolution: float
        The sidelength of each gridcell in meters. Default 1.
    validate: bool
        Whether to check the input graph for validity before conversion. 
        Default True. If false, the user is responsible for ensuring the graph is valid.
        For internal use only. 
    Returns
    -------
    p_c_graph : POINTER
        pointer to the created C FlowGrid structure.
    """

    if verbose:
        print("Converting DiGraph to FlowGrid_C...")
        print("--------------------------------------")

    # is a DAG
    if validate:
        if not isinstance(wrap, bool):
            raise TypeError(f"wrap must be a bool, got {type(wrap)}")
        if not isinstance(G, nx.DiGraph):
            raise TypeError(f"G must be a networkx.DiGraph, got {type(G)}")
        if not nx.is_directed_acyclic_graph(G):
            raise ValueError("Graph must be a DAG.")

        if verbose:
            print("\tGraph is a directed acyclic graph.")

    # pos attribute is valid
    pos_dict = nx.get_node_attributes(G, "pos")
    pos = list(pos_dict.values())
    pos_set = set(pos_dict.values())
    pos_dict_reversed = {v:k for k,v in pos_dict.items()}

    if validate:
        if len(pos) != len(G.nodes):
            raise ValueError("All graph nodes must have a 'pos' attribute.")
        if any(
            not isinstance(p, (tuple, list))  # Check if position is a tuple or list
            or len(p) != 2  # Check if it has exactly two elements
            or not all(isinstance(x, (int, np_integer)) for x in p)  # Check if both elements are integers
            or any(x < 0 for x in p)  # Check if both elements are non-negative
            for p in pos
        ):
            raise ValueError("All graph node 'pos' attributes must be non-negative (row:int, col:int) tuples.")
        
        if verbose:
            print("\tGraph 'pos' attributes are valid.")

    # spans a dense grid
    rows, cols = max(p[0] for p in pos) + 1, max(p[1] for p in pos) + 1

    if validate:
        if set(product(range(rows), range(cols))).difference(pos_set):
            raise ValueError(f"Graph does not cover a dense {rows}x{cols} grid.")
        if len(G.nodes) != rows * cols:
            raise ValueError(f"Graph does not cover a dense {rows}x{cols} grid (expected {rows*cols} nodes, got {len(G.nodes)}).")

        if verbose:
            print(f"\tGraph covers a dense {rows}x{cols} grid.")

        # is a spanning tree
        if any(G.out_degree(u) > 1 for u in G.nodes):
            raise ValueError("Graph must be a spanning tree (each node has out_degree <= 1).")
        roots = [u for u in G.nodes if G.out_degree(u) == 0]
        
        if verbose:
            print(f"\tFound {len(roots)} spanning trees.")

        # edges only connect to adjacent nodes (no skipping)
        for u, v in G.edges:
            r1, c1 = G.nodes[u]["pos"]
            r2, c2 = G.nodes[v]["pos"]
            dr, dc = abs(r2 - r1), abs(c2 - c1)
            if wrap:
                dr = min(dr, rows - dr)
                dc = min(dc, cols - dc)
            if max(dr, dc) != 1:
                raise ValueError(f"Edge ({u}->{v}) connects non-adjacent nodes at positions {(r1,c1)} and {(r2,c2)}.")
        
        if verbose:
            print("\tEdges connect only to adjacent nodes.")


    # helper function to get the the bit number to set in the edges attribute
    def direction_bit(pos1, pos2):
        r1, c1 = pos1
        r2, c2 = pos2
        dr, dc = r2 - r1, c2 - c1
        match dr, dc:
            case -1,  0: return 0  # N
            case -1,  1: return 1  # NE
            case  0,  1: return 2  # E
            case  1,  1: return 3  # SE
            case  1,  0: return 4  # S
            case  1, -1: return 5  # SW
            case  0, -1: return 6  # W
            case -1, -1: return 7  # NW
            case _: raise ValueError(f"Nodes at positions {pos1} and {pos2} are not adjacent.")
    # compute the drained area, adown, edges, downstream, and visited attributes for each node.
    # checking for crosses can come later: easier with edges defined.
    for n in nx.topological_sort(G):
        G.nodes[n]["visited"] = 0

        succs = list(G.successors(n))
        preds = list(G.predecessors(n))
        neighbors = succs + preds
        if len(neighbors) > 8:
            raise ValueError(f"Node {n} at position {G.nodes[n]['pos']} has {len(neighbors)} neighbors, but must have between 0 and 8.")
        if len(succs) > 1:
            raise ValueError(f"Node {n} at position {G.nodes[n]['pos']} has {len(succs)} successors, but must have at most 1.")

        G.nodes[n]["drained_area"] = resolution**2 + sum(G.nodes[p]["drained_area"] for p in preds)
        
        G.nodes[n]["edges"] = 0
        for nbr in neighbors:
            G.nodes[n]["edges"] |= (1 << direction_bit(G.nodes[n]["pos"], G.nodes[nbr]["pos"]))
        
        G.nodes[n]["downstream"] = _bindings.IS_ROOT
        G.nodes[n]["adown"] = rows * cols  # invalid index
        if len(succs):
            nsucc = succs[0]
            G.nodes[n]["downstream"] = direction_bit(G.nodes[n]["pos"], G.nodes[nsucc]["pos"])
            G.nodes[n]["adown"] = _bindings.libocn.fg_cart_to_lin(
                _bindings.CartPair_C(row=G.nodes[nsucc]["pos"][0], col=G.nodes[nsucc]["pos"][1]),
                _bindings.CartPair_C(row=rows, col=cols)
            )

    if verbose:
        print("\tComputed node attributes (drained_area, adown, edges, downstream, visited).")

    # check that edges do not cross each other
    #TODO: this is the current bottleneck of the conversion process. Consider profiling.
    if validate:
        for n in G.nodes:
            r, c = G.nodes[n]["pos"]
            
            down = G.nodes[n]["downstream"]
            if down % 2 == 0: continue  # Not a diagonal flow, cannot cross
            
            succs = list(G.successors(n))
            if len(succs) == 0: continue  # skip root node

            match down:
                case 1: r_check, c_check = r - 1, c      # NE flow: check N vertex
                case 7: r_check, c_check = r - 1, c      # NW flow: check N vertex
                case 3: r_check, c_check = r + 1, c      # SE flow: check S vertex
                case 5: r_check, c_check = r + 1, c      # SW flow: check S vertex
                case _: continue  # should have already been caught

            # find the node with pos (r_check, c_check)
            cross_check_node = pos_dict_reversed.get((r_check, c_check))
            if cross_check_node is None:
                raise ValueError(f"Node at position {(r_check, c_check)} does not exist!")
            cross_edges = G.nodes[cross_check_node]["edges"]
            if (
                G.nodes[n]["downstream"] == 1 and (cross_edges & (1 << 3))  # NE flow: N vertex has SE edge
                or (G.nodes[n]["downstream"] == 7 and (cross_edges & (1 << 5)))  # NW flow: N vertex has SW edge
                or (G.nodes[n]["downstream"] == 3 and (cross_edges & (1 << 1)))  # SE flow: S vertex has NE edge
                or (G.nodes[n]["downstream"] == 5 and (cross_edges & (1 << 7)))  # SW flow: S vertex has NW edge
            ):
                raise ValueError(f"Edge ({n}->{succs[0]}) crosses edge from node at position {(r_check, c_check)}.")

        if verbose:
            print("\tChecked for crossing edges.")

    # By now, the graph is validated and has all necessary attributes to create the C FlowGrid structure.
    p_c_graph = _bindings.libocn.fg_create_empty(_bindings.CartPair_C(row=rows, col=cols))
    if not p_c_graph:
        raise MemoryError("Failed to allocate memory for FlowGrid_C.")
    for n in G.nodes:
        r, c = G.nodes[n]["pos"]
        a = _bindings.libocn.fg_cart_to_lin(
            _bindings.CartPair_C(row=r, col=c),
            _bindings.CartPair_C(row=rows, col=cols)
        )
        v_c = _bindings.Vertex_C(
            drained_area=G.nodes[n]["drained_area"],
            adown=G.nodes[n]["adown"],
            edges=G.nodes[n]["edges"],
            downstream=G.nodes[n]["downstream"],
            visited=G.nodes[n]["visited"],
        )
        try:
            check_status(_bindings.libocn.fg_set_lin(p_c_graph, v_c, a))
        except Exception as e:
            _bindings.libocn.fg_destroy(p_c_graph)
            p_c_graph = None
            raise e
    
    p_c_graph.contents.resolution = float(resolution)
    p_c_graph.contents.nroots = len([n for n in G.nodes if G.out_degree(n) == 0])
    p_c_graph.contents.wrap = wrap
    
    # do not set energy

    if verbose:
        print("\tSuccessfully created FlowGrid_C from DiGraph.")

    return p_c_graph

def validate_digraph(dag:nx.DiGraph, verbose:bool=False) -> bool|str:
    """
    Validate the integrity of a FlowGrid.

    Parameters
    ----------
    dag : nx.DiGraph
        The directed acyclic graph to validate.
    """
    try:
        p_c_graph = from_digraph(dag, verbose=verbose)
        _bindings.libocn.fg_destroy(p_c_graph)
        p_c_graph = None
    except Exception as e:  # _digraph_to_flowgrid_c will destroy p_c_graph on failure
        return str(e)
    return "Graph is valid."

def validate_flowgrid(c_graph:_bindings.FlowGrid_C, verbose:bool=False) -> bool|str:
    """
    Validate the integrity of a FlowGrid_C structure.

    Parameters
    ----------
    c_graph : _bindings.FlowGrid_C
        The FlowGrid_C structure to validate.
    """
    try:
        dag = to_digraph(c_graph, verbose=verbose)
        p_c_graph = from_digraph(dag, verbose=verbose)
        _bindings.libocn.fg_destroy(p_c_graph)
        p_c_graph = None
    except Exception as e:  # _digraph_to_flowgrid_c will destroy p_c_graph on failure
        return str(e)
    return "Graph is valid."

__all__ = [
    "to_digraph",
    "from_digraph",
    "validate_digraph",
    "validate_flowgrid",
]
